# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader librairies

import collections

class RRReader(collections.deque):
    """ Round Robin reader, which are collections.deque with a
    readlines() method"""

    def readlines(self):
        """ return next line from queue """
        while 1:
            try:
                yield self.popleft()
            except IndexError:
                return
 



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader database connection handling
# COPY dichotomy on error

import os, sys, os.path, time, codecs, logging
from cStringIO import StringIO
from tempfile import gettempdir

from options import DRY_RUN, PEDANTIC, CLIENT_MIN_MESSAGES
from options import TRUNCATE, VACUUM
from options import INPUT_ENCODING, PG_CLIENT_ENCODING, DATESTYLE
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import PSYCOPG_VERSION
from options import PG_OPTIONS

from tools   import PGLoader_Error
from logger  import log

log.debug('Preferred psycopg version is %s' % PSYCOPG_VERSION)

if PSYCOPG_VERSION is None:
    # legacy import behavior
    log.debug('Trying psycopg2 then psycopg')
    try:
        import psycopg2.psycopg1 as psycopg
    except ImportError:
        log.info('No psycopg2 module found, trying psycopg1')

        try:
            import psycopg
        except ImportError, e:
            log.fatal('No psycopg module found')
            raise PGLoader_Error, e

elif PSYCOPG_VERSION == 1:
    try:
        log.info("Loading psycopg 1")
        import psycopg
    except ImportError, e:
        log.fatal("Can't load version %d of psycopg" % PSYCOPG_VERSION)
        raise PGLoader_Error, e
        
elif PSYCOPG_VERSION == 2:
    try:
        log.info("Loading psycopg 2")
        import psycopg2.psycopg1 as psycopg
    except ImportError, e:
        log.fatal("Can't load version %d of psycopg" % PSYCOPG_VERSION)
        raise PGLoader_Error, e

class db:
    """ a db connexion and utility class """
    def __init__(self,
                 dsn,
                 client_encoding = PG_CLIENT_ENCODING,
                 copy_every = 10000, commit_every = 1000, connect = True):
        """ Connects to the specified database """
        self.log     = log
        self.dbconn  = None
        self.dsn     = dsn

        # those parameters can be overwritten after db init
        # here's their default values
        self.copy_sep        = COPY_SEP
        self.copy_every      = copy_every
        self.commit_every    = commit_every
        self.null            = NULL
        self.empty_string    = EMPTY_STRING
        self.pg_options      = {}

        # this allows to specify configuration has columns = *
        # when true, we don't include column list in COPY statements
        self.all_cols = None

        if connect:
            self.reset()

    def __del__(self):
        """ db object destructor, we have to close the db connection """
        if self.dbconn is None:
            return

        if self.running_commands > 0:
            self.dbconn.commit()
            self.commits       += 1
            self.commited_rows += self.running_commands

        self.close()

    def close(self):
        """ close self.dbconn PostgreSQL connection """
        if self.dbconn is not None:
            try:
                self.log.info('closing current database connection')
            except IOError, e:
                # Ignore no space left on device...
                pass

            try:
                self.dbconn.close()
            except InterfaceError, e:
                # Ignore connection already closed
                pass
            self.dbconn = None

    def set_pg_options(self):
        """ set pg_options """
        for opt, val in self.pg_options.items():
            self.log.debug('Setting %s to %s', opt, val)
        
            sql = 'set session %s to %%s' % opt
            cursor = self.dbconn.cursor()
            try:
                cursor.execute(sql, [val])
            except (psycopg.ProgrammingError, psycopg.DataError), e:
                raise PGLoader_Error, e
            cursor.close()

    def get_all_columns(self, tablename):
        """ select the columns name list from catalog """

        if tablename.find('.') == -1:
            schemaname = 'public'
        else:
            try:
                schemaname, tablename = tablename.split('.')
            except ValueError, e:
                self.log.warning("db.get_all_columns: " + \
                                 "%s has more than one '.' separator" \
                                 % tablename)
                raise PGLoader_Error, e

        sql = """
  SELECT attname, attnum
    FROM pg_attribute
   WHERE attrelid = (SELECT oid
                       FROM pg_class
                      WHERE relname = %s
                            AND relnamespace = (SELECT oid
                                                  FROM pg_namespace
                                                 WHERE nspname = %s)
                    )
          AND attnum > 0 AND NOT attisdropped
ORDER BY attnum
"""
        self.log.debug("get_all_columns: %s %s %s" % (tablename, schemaname, sql))

        columns = []

        have_to_connect = self.dbconn is None
        if have_to_connect:
            self.reset()
            
        cursor  = self.dbconn.cursor()
        try:
            cursor.execute(sql, [tablename, schemaname])

            for row in cursor.fetchall():
                columns.append(row)

        except psycopg.ProgrammingError, e:
            raise PGLoader_Error, e

        cursor.close()

        if have_to_connect:
            self.close()

        return columns

    def reset(self):
        """ reset internal counters and open a new database connection """
        self.buffer            = None
        self.copy              = None # flag set to True when copy is called
        self.errors            = 0
        self.commits           = 0
        self.commited_rows     = 0
        self.running_commands  = 0
        self.last_commit_time  = time.time()
        self.first_commit_time = self.last_commit_time
        self.partial_coldef    = None

        try:
            self.close()
            self.log.debug('Debug: connecting to dns %s', self.dsn)

            self.dbconn = psycopg.connect(self.dsn)
            self.set_pg_options()
            
        except psycopg.OperationalError, e:
            # e.g. too many connections
            self.log.error(e)
            raise PGLoader_Error, "Can't connect to database"

    def print_stats(self):
        """ output some stats about recent activity """
        d = time.time() - self.first_commit_time
        u = self.commited_rows
        c = self.commits
        self.log.info(" %d rows copied in %d commits took %5.3f seconds", u, c, d)

        if self.errors > 0:
            self.log.error("%d database errors occured", self.errors)
            if self.copy and not VACUUM:
                self.log.info("Please VACUUM your database to recover space")
        else:
            if u > 0:
                self.log.info("No database error occured")
        return

    def is_null(self, value):
        """ return true if value is null, per configuration """
        return value == self.null

    def is_empty(self, value):
        """ return true if value is empty, per configuration """
        return value == self.empty_string

    def truncate(self, table):
        """ issue an SQL TRUNCATE TABLE on given table """
        if DRY_RUN:
            self.log.info("Won't truncate tables on dry-run mode")
            return
        
        sql = "TRUNCATE TABLE %s;" % table

        self.log.info('%s' % sql)
        
        try:
            cursor = self.dbconn.cursor()
            cursor.execute(sql)
            self.dbconn.commit()
        except Exception, error:
            self.log.error(error)
            raise PGLoader_Error, "Couldn't TRUNCATE table %s" % table
    
    def vacuum(self, table):
        """ issue VACUUM ANALYZE table """
        if DRY_RUN:
            self.log.info('no vacuum in dry-run mode')
            return -1

        sql = "VACUUM ANALYZE %s;" % table

        self.log.info('%s' % sql)
        
        try:
            cursor = self.dbconn.cursor()
            cursor.execute(sql)
            self.dbconn.commit()
        except Exception, error:
            self.log.error(error)
            raise PGLoader_Error, "Couldn't VACUUM table %s" % table

    def disable_triggers(self, table):
        """ issue ALTER TABLE table DISABLE TRIGGER ALL """
        if DRY_RUN:
            self.log.info("Won't disable triggers on dry-run mode")
            return
        
        sql = "ALTER TABLE %s DISABLE TRIGGER ALL;" % table

        self.log.info('%s' % sql)
        
        try:
            cursor = self.dbconn.cursor()
            cursor.execute(sql)
            self.dbconn.commit()
        except Exception, error:
            self.log.error(error)
            raise PGLoader_Error, "Couldn't DISABLE TRIGGERS on table %s" % table

    def enable_triggers(self, table):
        """ issue ALTER TABLE table ENABLE TRIGGER ALL """
        if DRY_RUN:
            self.log.info("Won't enable triggers on dry-run mode")
            return
        
        sql = "ALTER TABLE %s ENABLE TRIGGER ALL;" % table

        self.log.info('%s' % sql)
        
        try:
            cursor = self.dbconn.cursor()
            cursor.execute(sql)
            self.dbconn.commit()
        except Exception, error:
            self.log.error(error)
            raise PGLoader_Error, "Couldn't ENABLE TRIGGERS on table %s" % table

    
    def insert_blob(self, table, index, rowids,
                    blob_cname, data, btype,
                    input_line, reject):
        """ insert the given blob content into postgresql table

        return True on success, False on error
        """
        ok  = True
        sql = ""

        if btype == 'ifx_clob':
            data = data.replace("'", "\\'")
            sql = "UPDATE %s SET %s = %%s WHERE " % (table, blob_cname)

        elif btype == 'ifx_blob':
            data = data.tostring()
            sql = "UPDATE %s SET %s = %%s::bytea WHERE " % (table, blob_cname)

        values = [data]
        
        ##
        # Add a WHERE clause for each index
        first = True
        for name, col in index:
            if not first: sql  += " AND "
            else:         first = False
            
            sql += "%s = %%s" % name
            values.append(rowids[name])
        sql += ";"

        self.log.debug('%s' % sql)

        try:
            cursor = self.dbconn.cursor()
            cursor.execute(sql, values)

            # if execute raise an exception, don't count it as a
            # running command (waiting a commit)
            self.running_commands += 1

            if VERBOSE:
                str_rowids = ""
                for i,v in rowids.items():
                    if str_rowids != "": str_rowids += ", "
                    str_rowids += "%s:%s" % (i, v)
                self.log.debug('%s %s %s %6do', 
                          table, str_rowids, blob_cname, len(data))

            if self.running_commands == self.commit_every:
                now = time.time()
                self.dbconn.commit()
                
                self.commits += 1
                duration      = now - self.last_commit_time
                self.last_commit_time = now

                self.log.info("commit %d: %d updates in %5.3fs",
                         self.commits, self.running_commands, duration)

                self.commited_rows   += self.running_commands
                self.running_commands = 1

        except KeyboardInterrupt, error:
            # C-c was pressed, please stop processing
            self.dbconn.commit()
            raise PGLoader_Error, "Aborting on user demand (Interrupt)"

        except Exception, e:
            self.dbconn.commit()
            # don't use self.commited_rows here, it's only updated
            # after a commit
            self.log.error("update %d rejected: commiting (read log file %s)",
                      self.commits * self.commit_every + self.running_commands,
                      reject.reject_log)

            reject.log(str(e), input_line)
            self.errors += 1
            ok = False

        return ok

    def save_copy_buffer(self, tablename, debug = False):
        """ save copy buffer to a temporary file for further inspection """
        import tempfile
        (f, n) = tempfile.mkstemp(prefix='%s.' % tablename,
                                  suffix='.pgloader', dir=gettempdir())
        os.write(f, self.buffer.getvalue())
        os.close(f)
        if debug:
            self.log.debug("COPY data buffer saved in %s" % n)
        else:
            self.log.warning("COPY data buffer saved in %s" % n)
        return n

    def copy_from(self, table, columnlist,
                  columns, input_line, offsets,
                  reject, EOF = False):
        """ Generate some COPY SQL for PostgreSQL """
        ok = True
        if not self.copy: self.copy = True

        if EOF or self.running_commands == self.copy_every \
               and self.buffer is not None:
            # time to copy data to PostgreSQL table

            if self.buffer is None:
                self.log.warning("no data to COPY")
                return False
            
            ##
            # build the table colomns specs from parameters
            # ie. we always issue COPY table (col1, col2, ..., coln) commands
            tablename = table
            if self.all_cols:
                table = table
            else:
                table = "%s (%s) " % (table, ", ".join(columnlist))

            self.log.debug("COPY will use table definition: '%s'" % table)
                
            if CLIENT_MIN_MESSAGES <= logging.DEBUG:
                self.save_copy_buffer(tablename, debug = True)

            self.buffer.seek(0)
            now = time.time()
                
            try:
                cursor = self.dbconn.cursor()
                r = self.cursor_copy_from(cursor, self.buffer, table, self.copy_sep)
                self.dbconn.commit()

                self.commits         += 1
                duration              = now - self.last_commit_time
                self.last_commit_time = now

                self.log.info("COPY %d: %d rows copied in %5.3fs",
                         self.commits, self.running_commands, duration)

                # prepare next run
                self.buffer.close()
                self.buffer = None
                self.commited_rows   += self.running_commands
                self.running_commands = 0

            except (psycopg.ProgrammingError,
                    psycopg.DatabaseError), error:
                # rollback current transaction
                self.dbconn.rollback()

                self.log.warning('COPY error, trying to find on which line')
                if CLIENT_MIN_MESSAGES > logging.DEBUG:
                    # in DEBUG mode, copy buffer has already been saved
                    # to file
                    self.save_copy_buffer(tablename)

                # copy recovery process
                now = time.time()
                c, ok, ko = self.copy_from_buff(table, 
                                                self.buffer, 
                                                self.first_offsets,
                                                self.running_commands, 
                                                reject)

                duration              = now - self.last_commit_time
                self.commits         += c
                self.last_commit_time = now
                self.commited_rows   += ok
                self.errors          += ko

                self.log.warning('COPY error recovery done (%d/%d) in %5.3fs',
                            ko, ok, duration)

                # commit this transaction
                self.dbconn.commit()

                # recovery process has closed the buffer
                self.buffer = None
                self.running_commands = 0

        # prepare next run
        if self.buffer is None:
            self.buffer = StringIO()
            self.first_offsets = offsets

        self.prepare_copy_data(columns, input_line, reject)
        self.running_commands += 1
        return ok

    def copy_from_buff(self, table, buff, first_offsets, count, reject):
        """ If copy returned an error, try to detect wrong input line(s) """

        if count == 1:
            msg = self.copy_error_message(first_offsets, 0)
            reject.log(msg, buff.getvalue())
            buff.close()
            self.log.debug('found one more line in error')

            # returns commits, ok, ko
            return 0, 0, 1
        
        ##
        # Dichotomy
        # we cut the buffer into two buffers, try to copy from them
        a = StringIO()
        b = StringIO()
        n = 0
        m = count / 2

        # return values, copied lines and errors
        commits = ok = ko = 0 

        buff.seek(0)
        for line in buff.readlines():
            if n < m:
                a.write(line)
            else:
                b.write(line)
            n += 1

        # we don't need no more orgininal buff
        buff.close()

        self.log.debug('Trying to find errors, dividing %d lines in %d and %d',
                  count, m, n-m)

        # now we have two buffers to copy to PostgreSQL database
        cursor = self.dbconn.cursor()
        for (x, xcount) in [(a, m), (b, n-m)]:
            try:
                x.seek(0)
                self.cursor_copy_from(cursor, x, table, self.copy_sep)
                self.dbconn.commit()
                x.close()

                self.log.debug("COPY ERROR handling progress: %d rows copied",
                          xcount)

                x.close()
                commits += 1
                ok += xcount

            except Exception, error:
                self.dbconn.commit()
                
                # if a is only one line long, reject this line
                if xcount == 1:
                    ko += 1

                    linecount = 0
                    if x == b:
                        linecount += m

                    msg = self.copy_error_message(first_offsets, linecount)
                    msg += '\nCOPY error: %s' % error
                    reject.log(msg, x.getvalue())
                    self.log.debug('Notice: found one more line in error')
                    self.log.debug(x.getvalue())

                else:
                    new_offsets = first_offsets
                    if x == b:
                        new_offsets = [m + o for o in first_offsets]

                    _c, _o, _k = self.copy_from_buff(table, 
                                                     x,
                                                     new_offsets,
                                                     xcount, reject)
                    commits += _c
                    ok += _o
                    ko += _k

        return commits, ok, ko

    def cursor_copy_from(self, cursor, buffer, table, delimiter):
        """ call psycopg copy command, in expert mode if available """
        
        if hasattr(cursor, 'copy_expert'):
            self.log.debug("using copy_expert")
            sql = "COPY %s FROM STDOUT WITH DELIMITER '%s'" % (table, delimiter)
            return cursor.copy_expert(sql, buffer)
        else:
            return cursor.copy_from(buffer, table, delimiter)


    def prepare_copy_data(self, columns, input_line, reject):
        """ add a data line to copy buffer """
        if columns is not None:
            first_col = True

            for c in columns:
                # default text format COPY delimiter
                if not first_col: self.buffer.write(self.copy_sep)
                else:             first_col = False

                if self.is_null(c):
                    # null column value: \N
                    self.buffer.write('\N')

                elif self.is_empty(c):
                    # empty string has been read
                    self.log.debug("empty string read: '%s'" % c)
                    self.buffer.write('')

                else:
                    # for a list of chars to replace, please have a look to
                    # http://www.postgresql.org/docs/8.1/static/sql-copy.html
                    if INPUT_ENCODING is not None:
                        try:
                            c = c.encode(INPUT_ENCODING)
                        except UnicodeDecodeError, e:
                            reject.log(['Codec error', str(e)], input_line)

                    # in _split_line we remove delimiter escaping
                    # in order for backslash escaping not to de-escape it
                    # we then have to escape delimiters explicitely now
                    for orig, escaped in [('\\', '\\\\'),
                                          (self.copy_sep,
                                           '\\%s' % self.copy_sep),
                                          ('\b', '\\b'),
                                          ('\f', '\\f'),
                                          ('\n', '\\n'),
                                          ('\r', '\\r'),
                                          ('\t', '\\t'),
                                          ('\v', '\\v')]:
                        try:
                            c = c.replace(orig, escaped)
                        except TypeError, e:
                            self.log.error("db.prepare_copy_data columns %s"    % str(columns))
                            self.log.error("db.prepare_copy_data input_line %s" % str(input_line))
                            self.log.error("TypeError: '%s'.replace(%s, %s)"    % (c, orig, escaped))
                            raise PGLoader_Error, e

                    self.buffer.write(c)

            # end of row, \n
            self.buffer.write('\n')


    def copy_error_message(self, offsets, error_buff_offset):
        """ Build the COPY pgloader error message with line numbers """
        msg = 'COPY error on line'
        if type(offsets) == type([]):
            if len(offsets) > 1:
                msg += 's'
            msg += 's %s' % ' '.join([str(x + error_buff_offset)
                                      for x in offsets])
                
        else:
            # offsets is (start position (byte, ftell()), [line, numbers])
            msg += ' '

        return msg



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader text format reader
#
# handles configuration, parse data, then pass them to database module for
# COPY preparation

import os, sys, os.path, time, csv
from cStringIO import StringIO

from tools    import PGLoader_Error, Reject, parse_config_string
from db       import db
from lo       import ifx_clob, ifx_blob
from reader   import DataReader, UnbufferedFileReader

from options import DRY_RUN, PEDANTIC
from options import TRUNCATE, VACUUM
from options import COUNT, FROM_COUNT, FROM_ID
from options import INPUT_ENCODING, PG_CLIENT_ENCODING
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import NEWLINE_ESCAPES

class CSVReader(DataReader):
    """
    Read some CSV formatted data
    """

    def readconfig(self, config, name, template):
        """ get this reader module configuration from config file """
        DataReader.readconfig(self, config, name, template)

        self._getopt('doublequote', config, name, template, True)
        if self.doublequote is not True:
            self.doublequote = self.doublequote == 'True'
        
        self._getopt('escapechar', config, name, template, None)
        if self.escapechar is not None:
            self.escapechar = self.escapechar[0]

        self._getopt('quotechar', config, name, template, '"')
        self.quotechar = self.quotechar[0]

        self._getopt('skipinitialspace', config, name, template, False)
        if self.skipinitialspace is not False:
            self.skipinitialspace = self.skipinitialspace == 'True'

        self._getopt('field_size_limit', config, name, template, -1, "mem")

        for opt in ['doublequote', 'escapechar',
                    'quotechar', 'skipinitialspace', 'field_size_limit']:
            
            self.log.debug("reader.readconfig %s: '%s'" \
                           % (opt, self.__dict__[opt]))

    def readlines(self):
        """ read data from configured file, and generate (yields) for
        each data line: line, columns and rowid """

        # make a dialect, then implement a reader with it
        class pgloader_dialect(csv.Dialect):
            delimiter        = self.field_sep
            doublequote      = self.doublequote
            escapechar       = self.escapechar
            quotechar        = self.quotechar
            skipinitialspace = self.skipinitialspace

            lineterminator   = '\r\n'
            quoting          = csv.QUOTE_MINIMAL
            
        csv.register_dialect('pgloader', pgloader_dialect)

        self.fd = UnbufferedFileReader(self.filename, self.log,
                                       encoding = self.input_encoding,
                                       start    = self.start,
                                       end      = self.end,
                                       skip_head_lines = self.skip_head_lines)
        
        # don't forget COUNT and FROM_COUNT option in CSV mode
        nb_lines     = self.skip_head_lines
        begin_linenb = None
        last_line_nb = 1

        # set the field_size_limit, from python 2.5
        if self.field_size_limit != -1:
            try:
                csv.field_size_limit(self.field_size_limit)
                self.log.debug("csv.field_size_limit(%d)" \
                               % self.field_size_limit)
            except AttributeError:
                #'module' object has no attribute 'field_size_limit'
                self.log.warning("field_size_limit is new in python version 2.5")
                pass

        # now read the lines
        for columns in csv.reader(self.fd, dialect = 'pgloader'):
            # we count logical lines
            nb_lines += 1

            line         = self.field_sep.join(columns)
            offsets      = range(last_line_nb, self.fd.line_nb)
            last_line_nb = self.fd.line_nb
            
            if self.start:
                offsets = (self.start, offsets)

            yield offsets, line, columns
            
        return



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader Large Object support

from cStringIO import StringIO
from tools     import PGLoader_Error
from options   import DRY_RUN, PEDANTIC
from options   import INPUT_ENCODING

class ifx_lo:
    """ an Informix Large Object data file as given by UNLOAD """

    def __del__(self):
        """ close self.file on object destruction """
        self.file.close()

    def extract(self, rowid, field_nb, begin, length):
        """ extract given positionned data from Informix Clob out file, save
        them in a file and returns the filename where text is stored """

        # get stream position and data length
        begin  = long(begin, 16)
        length = long(length, 16)

        # get blob data
        self.file.seek(begin)
        try:
            content = self.file.read(length)
        except UnicodeDecodeError, e:
            # as of now, this is a fatal error
            print
            print 'Fatal error in clob file %s at position Ox%x' \
                  % (self.filename, begin)
            raise PGImport_Error, e

        return content

class ifx_clob(ifx_lo):
    """ Informix Text Large Object file """

    def __init__(self, log, filename, input_encoding):
        """ init a clob object  """
        self.log       = log
        self.file      = None
        self.filename  = filename

        if self.file is None:
            if input_encoding is not None:
                import codecs
                self.file = codecs.open(self.filename, 'r',
                                        encoding = input_encoding)
            else:
                self.file = open(self.filename, 'r')

            self.log.info("Opening informix clob file: %s", self.filename)
                
class ifx_blob(ifx_lo):
    """ Informix Binary Large Object file """

    def __init__(self, log, filename, field_sep):
        """ init a clob object  """
        self.log       = log
        self.file      = None
        self.filename  = filename
        self.field_sep = field_sep # used by bytea_escape

        # some helpers for bytea escaping
        self.octals  = range(0, 32)
        self.octals += range(127, 256)

        if self.file is None:
            self.file = open(self.filename, 'rb')
            self.log.info("Opening informix blob file: %s", self.filename)

    def bytea_escape(self, bitstring):
        """ escape chars from bitstring for PostgreSQL bytea input
        see http://www.postgresql.org/docs/8.1/static/datatype-binary.html
        """
        escaped = StringIO()
        bsize   = len(bitstring)
        pos     = 0

        while pos < bsize:
            c = bitstring[pos]
            o = ord(c)

            if o in self.octals or c == self.field_sep:
                # PostgreSQL wants octal numbers!
                escaped.write('\\%03o' % o)
            elif o == 39:
                escaped.write('\\047')
            elif o == 92:
                escaped.write('\\134')
            else:
                escaped.write(c)

            pos += 1

        r = escaped.getvalue()
        escaped.close()
        
        return r

    def extract(self, rowid, field_nb, begin, length):
        """ extract content, then bytea escape it """

        content = ifx_lo.extract(self, rowid, field_nb, begin, length)
        return self.bytea_escape(content)
            



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader logging facility
#
# standard error levels are used for code and configuration error messages
# data error logging is managed by tools.Reject class

from tools import PGLoader_Error
import logging
from tempfile import gettempdir
import os

def init(client_min_messages = logging.INFO,
         log_min_messages = logging.DEBUG,
	 filename = os.path.join (gettempdir(), 'pgloader.log')):
    """ set the console logging """

    fmt = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'

    try:
        logging.basicConfig(level    = log_min_messages,
                            format   = fmt,
                            datefmt  = '%d-%m-%Y %H:%M:%S',
                            filename = filename,
                            filemode = 'w')
    except TypeError:
        # very old python (2.3 or such) didn't have kwargs in basicConfig
        logfile = logging.FileHandler(filename, filemode)
        logfile.setLevel(log_min_messages)
        logfile_fmt = logging.Formatter(fmt, datefmt)
        logfile.setFormatter(logfile_fmt)
        
        logging.getLogger('').addHandler(logfile)
    except IOError, e:
        raise PGLoader_Error, e

    console = logging.StreamHandler()
    console.setLevel(client_min_messages)

    formatter = logging.Formatter('%(name)-12s %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logging.getLogger('').addHandler(console)

    return logging.getLogger('pgloader')

def getLogger(name):
    """ return a new logger instance named name, properly configured """
    return logging.getLogger(name)

def level(name):
    """ return a logging level from user string """

    if name.upper() == 'DEBUG':
        return logging.DEBUG

    elif name.upper() == 'INFO':
        return logging.INFO

    elif name.upper() == 'WARNING':
        return logging.WARNING

    elif name.upper() == 'ERROR':
        return logging.ERROR

    elif name.upper() == 'CRITICAL':
        return logging.CRITICAL

    else:
        return logging.NOTSET



#! /usr/bin/env python
# Author: Dimitri Fontaine <dim@tapoueh.org>

"""
PostgreSQL data import tool, see included man page.
"""

import os, sys, os.path, time, codecs, logging, threading, signal
from cStringIO import StringIO
from tempfile import gettempdir

import pgloader.options
import pgloader.tools
import pgloader.logger
from pgloader.tools import PGLoader_Error

def parse_options():
    """ Parse given options """
    import ConfigParser
    from optparse import OptionParser

    usage  = "%prog [-c <config_filename>] Section [Section ...]"
    parser = OptionParser(usage = usage)
    
    parser.add_option("--version", action = "store_true",
                      dest    = "version",
                      default = False,
                      help    = "show pgloader version")

    parser.add_option("-c", "--config", dest = "config",
                      default = "pgloader.conf",
                      help    = "configuration file, defauts to pgloader.conf")

    parser.add_option("-p", "--pedantic", action = "store_true",
                      dest    = "pedantic",
                      default = False,
                      help    = "pedantic mode, stop processing on warning")

    parser.add_option("-d", "--debug", action = "store_true",
                      dest    = "debug",
                      default = False,
                      help    = "add some debug information (a lot of)")

    parser.add_option("-v", "--verbose", action = "store_true",
                      dest    = "verbose",
                      default = False,
                      help    = "be verbose and about processing progress")

    parser.add_option("-q", "--quiet", action = "store_true",
                      dest    = "quiet",
                      default = False,
                      help    = "be quiet, only print out errors")

    parser.add_option("-l", "--level", dest = "loglevel",
                      default = None,
                      help    = "loglevel to use: ERROR, WARNING, INFO, DEBUG")

    parser.add_option("-L", "--logfile", dest = "logfile",
                      default = os.path.join (gettempdir(), "pgloader.log"),
                      help    = "log file, defauts to $TMPDIR/pgloader.log")

    parser.add_option("-r", "--reject-log", dest = "reject_log",
                      default = None,
                      help    = "log file for rejected data error messages")

    parser.add_option("-j", "--reject-data", dest = "reject_data",
                      default = None,
                      help    = "log file for rejected data, bad input data")

    parser.add_option("-s", "--summary", action = "store_true",
                      dest    = "summary",
                      default = False,
                      help    = "print a summary")

    parser.add_option("-n", "--dry-run", action = "store_true",
                      dest    = "dryrun",
                      default = False,
                      help    = "simulate operations, don't connect to the db")

    parser.add_option("-T", "--truncate", action = "store_true",
                      dest    = "truncate",
                      default = False,
                      help    = "truncate tables before importing data")

    parser.add_option("-D", "--disable-triggers", action = "store_true",
                      dest    = "triggers",
                      default = False,
                      help    = "Disable triggers before loading, Enable them again after")

    parser.add_option("-V", "--vacuum", action = "store_true",
                      dest = "vacuum",
                      default = False,
                      help    = "vacuum tables after data loading")

    parser.add_option("-C", "--count", dest = "count",
                      default = None, type = "int",
                      help    = "number of input lines to process")
    
    parser.add_option("-F", "--from", dest = "fromcount",
                      default = 0, type = "int",
                      help    = "number of input lines to skip")

    parser.add_option("-I", "--from-id", dest = "fromid",
                      default = None,
                      help    = "wait for given id on input to begin")

    parser.add_option("-f", "--field-separator", dest = "fsep",
                      default = pgloader.options.FIELD_SEP,
                      help    = "default field separator to use")

    parser.add_option("-E", "--encoding", dest = "encoding",
                      default = None,
                      help    = "input files encoding")

    parser.add_option("-o", "--pg-options", dest = "pg_options", action = "append",
                      help    = "list of PostgreSQL options you want to SET")

    parser.add_option("-t", "--section-threads", dest = "section_threads",
                      default = pgloader.options.SECTION_THREADS, 
                      type    = "int",
                      help    = "number of threads used per sections, default is 1")

    parser.add_option("-m", "--max-parallel-sections", dest = "parallel",
                      default = pgloader.options.MAX_PARALLEL_SECTIONS,
                      type    = "int",
                      help    = "number of sections to load in parralel, " +\
                          "default is 1")

    parser.add_option("-R", "--reformat_path", dest = "reformat_path",
                      default = None,
                      help    = "PATH where to find reformat python modules")

    parser.add_option("-1", "--psycopg1", action = "store_true",
                      dest    = "psycopg1",
                      default = False,
                      help    = "Force usage of psycopg1")

    parser.add_option("-2", "--psycopg2", action = "store_true",
                      dest    = "psycopg2",
                      default = False,
                      help    = "Force usage of psycopg2")

    parser.add_option("--psycopg-version", dest = "psycopg_version",
                      default = None,
                      help    = "Force usage of given version of psycopg")

    parser.add_option("--load-from-stdin", action = "store_true",
                      dest = "stdin",
                      default = False,
                      help    = "Load standard input data into given table name")

    parser.add_option("--load-to-table", dest = "table",
                      default = None,
                      help    = "Load to given table when --load-from-stdin")

    parser.add_option("--boundaries", dest = "boundaries",
                      default = None,
                      help    = "Load only in given boundaries, Start..End")

    (opts, args) = parser.parse_args()

    if opts.version:
        print "PGLoader version %s" % pgloader.options.PGLOADER_VERSION
        sys.exit(0)

    # check existence and read ability of config file
    if not os.path.exists(opts.config):
        print >>sys.stderr, \
              "Error: Configuration file %s does not exists" % opts.config
        print >>sys.stderr, parser.format_help()
        sys.exit(1)

    if not os.access(opts.config, os.R_OK):
        print >>sys.stderr, \
              "Error: Can't read configuration file %s" % opts.config
        print >>sys.stderr, parser.format_help()
        sys.exit(1)

    if opts.fromcount != 0 and opts.fromid is not None:
        print >>sys.stderr, \
              "Error: Can't set both options fromcount (-F) AND fromid (-I)"
        sys.exit(1)

    if opts.quiet and (opts.verbose or opts.debug):
        print >>sys.stderr, \
              "Error: Can't be verbose and quiet at the same time!"
        sys.exit(1)

    psyco_opts = 0
    for opt in [opts.psycopg1, opts.psycopg2, opts.psycopg_version]:
        if opt:
            psyco_opts += 1

    if psyco_opts > 1:
        print >>sys.stderr, \
              "Error: please use only one of the psycopg options"
        sys.exit(1)

    if opts.psycopg_version is not None:
        if opts.psycopg_version in ("1", "2"):
            opts.psycopg_version = int(opts.psycopg_version)

        else:
            print >>sys.stderr, \
                  "Error: psycopg_version can only be set to either 1 or 2"
            sys.exit(1)
        
    # if debug, then verbose
    if opts.debug:
        opts.verbose = True

    pgloader.options.DRY_RUN    = opts.dryrun
    pgloader.options.DEBUG      = opts.debug
    pgloader.options.VERBOSE    = opts.verbose
    pgloader.options.QUIET      = opts.quiet
    pgloader.options.SUMMARY    = opts.summary    
    pgloader.options.PEDANTIC   = opts.pedantic

    pgloader.options.TRUNCATE   = opts.truncate
    pgloader.options.VACUUM     = opts.vacuum
    pgloader.options.TRIGGERS   = opts.triggers
    
    pgloader.options.COUNT      = opts.count
    pgloader.options.FROM_COUNT = opts.fromcount
    pgloader.options.FROM_ID    = opts.fromid
    pgloader.options.FIELD_SEP  = opts.fsep

    pgloader.options.SECTION_THREADS       = opts.section_threads
    pgloader.options.MAX_PARALLEL_SECTIONS = opts.parallel

    pgloader.options.LOAD_FROM_STDIN = opts.stdin
    pgloader.options.LOAD_TO_TABLE   = opts.table

    if pgloader.options.MAX_PARALLEL_SECTIONS is None:
        from pgloader.options import DEFAULT_MAX_PARALLEL_SECTIONS
        pgloader.options.MAX_PARALLEL_SECTIONS = DEFAULT_MAX_PARALLEL_SECTIONS

    pgloader.options.INPUT_ENCODING = opts.encoding

    if opts.reformat_path:
        pgloader.options.REFORMAT_PATH = opts.reformat_path

    pgloader.options.LOG_FILE = opts.logfile

    # reject file names must contain one %s which will get replaced by the
    # section name
    if opts.reject_log:
        try:
            unused = opts.reject_log % "that would be a section name"
        except TypeError, e:
            # TypeError: not all arguments converted during string formatting
            # TypeError: not enough arguments for format string
            print >>sys.stderr, \
                "Error: reject log must contain a '%s' place holder for section name"
            sys.exit(1)
        pgloader.options.REJECT_LOG_FILE = opts.reject_log

    if opts.reject_data:
        try:
            unused = opts.reject_data % "that would be a section name"
        except TypeError, e:
            # TypeError: not all arguments converted during string formatting
            # TypeError: not enough arguments for format string
            print >>sys.stderr, \
                "Error: reject data must contain a '%s' place holder for section name"
            sys.exit(1)
        pgloader.options.REJECT_DATA_FILE = opts.reject_data

    if opts.loglevel:
        loglevel = pgloader.logger.level(opts.loglevel)
        pgloader.options.CLIENT_MIN_MESSAGES = loglevel
    elif opts.debug:
        pgloader.options.CLIENT_MIN_MESSAGES = logging.DEBUG
    elif opts.verbose:
        pgloader.options.CLIENT_MIN_MESSAGES = logging.INFO
    elif opts.quiet:
        pgloader.options.CLIENT_MIN_MESSAGES = logging.ERROR

    if opts.pg_options:
        pgloader.options.PG_OPTIONS = {}
        for o in opts.pg_options:
            try:
                n, v = [x.strip() for x in o.split('=')]
                if v == "":
                    raise ValueError
                pgloader.options.PG_OPTIONS[n] = v
            except ValueError, e:
                print >>sys.stderr, \
                    "Error: PostgreSQL options must have the form 'name=value'"
                sys.exit(1)

    if opts.psycopg1:
        pgloader.options.PSYCOPG_VERSION = 1
    elif opts.psycopg2:
        pgloader.options.PSYCOPG_VERSION = 2
    else:
        pgloader.options.PSYCOPG_VERSION = opts.psycopg_version

    if opts.boundaries:
        if opts.stdin:
            print >>sys.stderr, \
                  "Error: You can't set the boundaries of stdin."
            sys.exit(1)

        try:
            start, end = [int(x) for x in opts.boundaries.split("..")]
            pgloader.options.FILE_BOUNDARIES = (start, end)
        except ValueError, e:
            print >>sys.stderr, \
                  "Error: boundaries should be an integer range written X..Y"
            sys.exit(1)

    return opts.config, args

def parse_config(conffile):
    """ Parse the configuration file """
    section = 'pgsql'

    # Now read pgsql configuration section
    import ConfigParser
    config = ConfigParser.ConfigParser()

    try:
        config.read(conffile)
    except:
        print >>sys.stderr, "Error: Given file is not a configuration file"
        sys.exit(4)

    if not config.has_section(section):
        print >>sys.stderr, "Error: Please provide a [%s] section" % section
        sys.exit(5)

    # load some options
    # this has to be done after command line parsing
    from pgloader.options  import DRY_RUN, VERBOSE, DEBUG, PEDANTIC
    from pgloader.options  import NULL, EMPTY_STRING
    from pgloader.options  import CLIENT_MIN_MESSAGES, LOG_FILE
    from pgloader.options  import PG_OPTIONS
    from pgloader.tools    import check_dirname

    # first read the logging configuration
    if not CLIENT_MIN_MESSAGES:
        if config.has_option(section, 'client_min_messages'):
            cmm = config.get(section, 'client_min_messages')
            pgloader.options.CLIENT_MIN_MESSAGES = pgloader.logger.level(cmm)
        else:
            # CLIENT_MIN_MESSAGES has not been set at all
            pgloader.options.CLIENT_MIN_MESSAGES = logging.INFO

    if config.has_option(section, 'log_min_messages'):
        lmm = config.get(section, 'log_min_messages')
        pgloader.options.LOG_MIN_MESSAGES = pgloader.logger.level(lmm)
    else:
        pgloader.options.LOG_MIN_MESSAGES = logging.INFO

    if config.has_option(section, 'log_file'):
        # don't overload the command line -L option if given
        if not pgloader.options.LOG_FILE:
            pgloader.options.LOG_FILE = config.get(section, 'log_file')

    if pgloader.options.LOG_FILE:
        ok, logdir_mesg = check_dirname(pgloader.options.LOG_FILE)
        if not ok:
            # force default setting
            pgloader.options.LOG_FILE = pgloader.options.DEFAULT_LOG_FILE

    try:
        log = pgloader.logger.init(pgloader.options.CLIENT_MIN_MESSAGES,
                                   pgloader.options.LOG_MIN_MESSAGES,
                                   pgloader.options.LOG_FILE)
    except PGLoader_Error, e:
        try:
            log = pgloader.logger.init(pgloader.options.CLIENT_MIN_MESSAGES,
                                       pgloader.options.LOG_MIN_MESSAGES,
                                       pgloader.options.DEFAULT_LOG_FILE)

            log.warning(e)
            log.warning("Using default logfile %s",
                        pgloader.options.DEFAULT_LOG_FILE)
        except PGLoader_Error, e:
            print e
            sys.exit(8)
        
    pgloader.logger.log = log

    log.info("Logger initialized")
    if logdir_mesg:
        log.error(logdir_mesg)
        log.error("Default logfile %s has been used instead",
                  pgloader.options.LOG_FILE)

    if config.has_option(section, 'input_encoding'):
        input_encoding = pgloader.tools.parse_config_string(
            config.get(section, 'input_encoding'))
        pgloader.options.INPUT_ENCODING = input_encoding

    # optionnal global newline_escapes
    if config.has_option(section, 'newline_escapes'):
        setting = pgloader.tools.parse_config_string(
            config.get(section, 'newline_escapes'))
        pgloader.options.NEWLINE_ESCAPES = setting

    # Then there are null and empty_string optionnal parameters
    # They canbe overriden in specific table configuration
    if config.has_option(section, 'null'):
        pgloader.options.NULL = pgloader.tools.parse_config_string(
            config.get(section, 'null'))

    if config.has_option(section, 'empty_string'):
        pgloader.options.EMPTY_STRING = pgloader.tools.parse_config_string(
            config.get(section, 'empty_string'))

    if config.has_option(section, 'reformat_path'):
        # command line value is prefered to config format one
        if not pgloader.options.REFORMAT_PATH:
            rpath = config.get(section, 'reformat_path')
            pgloader.options.REFORMAT_PATH = rpath

    if config.has_option(section, 'max_parallel_sections'):
        if not pgloader.options.MAX_PARALLEL_SECTIONS:
            mps = config.getint(section, 'max_parallel_sections')
            pgloader.options.MAX_PARALLEL_SECTIONS = mps

    return config

def myprint(l, line_prefix = "  ", cols = 78):
    """ pretty print list l elements """
    # some code for pretty print
    lines = []
    
    tmp = line_prefix
    for e in l:
        if len(tmp) + len(e) > cols:
            lines.append(tmp)
            tmp = line_prefix
            
        if tmp != line_prefix: tmp += " "
        tmp += e
        
    lines.append(tmp)

    return lines

def duration_pprint(duration):
    """ pretty print duration (human readable information) """
    if duration > 3600:
        h  = int(duration / 3600)
        m  = int((duration - 3600 * h) / 60)
        s  = duration - 3600 * h - 60 * m + 0.5
        return '%2dh%02dm%03.1f' % (h, m, s)
    
    elif duration > 60:
        m  = int(duration / 60)
        s  = duration - 60 * m
        return ' %02dm%06.3f' % (m, s)
        
    else:
        return '%10.3f' % duration

def print_summary(dbconn, sections, summary, td):
    """ print a pretty summary """
    from pgloader.options  import VERBOSE, DEBUG, QUIET, SUMMARY
    from pgloader.options  import DRY_RUN, PEDANTIC, VACUUM
    from pgloader.pgloader import PGLoader

    retcode = 0

    t= 'Table name        |    duration |    size |  copy rows |     errors '
    _= '===================================================================='

    tu = te = ts = 0 # total updates, errors, size

    if False and not DRY_RUN:
        dbconn.reset()
        cursor = dbconn.dbconn.cursor()

    s_ok = 0
    for s in sections:
        if s not in summary:
            continue

        s_ok += 1
        if s_ok == 1:
            # print pretty sumary header now
            print
            print t
            print _

        if summary[s]:
            t, d, u, e = summary[s]
            d = duration_pprint(d)
        else:
            t = s
            d = '%9s ' % '-'
            u = e = 0

        if False and not DRY_RUN:
            sql = "select pg_total_relation_size(%s), " + \
                  "pg_size_pretty(pg_total_relation_size(%s));"
            cursor.execute(sql, [t, t])
            octets, sp = cursor.fetchone()
            ts += octets

            if sp[5:] == 'bytes': sp = sp[:-5] + ' B'
        else:
            sp = '-'

        tn = s
        if len(tn) > 18:
            tn = s[0:15] + "..."

        print '%-18s| %ss | %7s | %10d | %10d' % (tn, d, sp, u, e)

        tu += u
        te += e

        if e > 0:
            retcode += 1

    if s_ok > 1:
        td = duration_pprint(td)

        # pretty size
        if False and not DRY_RUN:
            cursor.execute("select pg_size_pretty(%s);", [ts])
            [ts] = cursor.fetchone()
            if ts[5:] == 'bytes': ts = ts[:-5] + ' B'
        else:
            ts = '-'

        print _
        print 'Total             | %ss | %7s | %10d | %10d' \
              % (td, ts, tu, te)

        if False and not DRY_RUN:
            cursor.close()

    return retcode

def load_data():
    """ read option line and configuration file, then process data
    import of given section, or all sections if no section is given on
    command line """

    # first parse command line options, and set pgloader.options values
    # accordingly
    conffile, args = parse_options()

    # now init db connection
    config = parse_config(conffile)

    from pgloader.logger  import log
    from pgloader.tools   import read_path, check_path
    from pgloader.options import VERBOSE
    
    import pgloader.options
    if pgloader.options.REFORMAT_PATH:
        rpath  = read_path(pgloader.options.REFORMAT_PATH, log, check = False)
        crpath = check_path(rpath, log)
    else:
        rpath  = crpath  = None

    if not crpath:
        if rpath:
            # don't check same path entries twice
        
            default_rpath = set(crpath) \
                            - set(pgloader.options.DEFAULT_REFORMAT_PATH)
        else:
            default_rpath = pgloader.options.DEFAULT_REFORMAT_PATH
        
        pgloader.options.REFORMAT_PATH = check_path(default_rpath, log)
    else:
        pgloader.options.REFORMAT_PATH = rpath

    log.info('Reformat path is %s', pgloader.options.REFORMAT_PATH)

    # load some pgloader package modules
    from pgloader.options  import VERBOSE, DEBUG, QUIET, SUMMARY
    from pgloader.options  import DRY_RUN, PEDANTIC, VACUUM
    from pgloader.options  import MAX_PARALLEL_SECTIONS
    from pgloader.options  import LOAD_FROM_STDIN, LOAD_TO_TABLE
    from pgloader.options  import FILE_BOUNDARIES
    from pgloader.pgloader import PGLoader
    from pgloader.tools    import PGLoader_Error

    sections = []
    summary  = {}

    # args are meant to be configuration sections, or filenames, or stdin
    if LOAD_FROM_STDIN:
        if FILE_BOUNDARIES is not None:
            log.warning("Can't use --boundaries on stdin")

        if len(args) == 0:
            s = '<stdin>'
            config.add_section(s)
            config.set(s, 'table', LOAD_TO_TABLE)
            config.set(s, 'filename', 'sys.stdin')
            config.set(s, 'columns', '*')
            config.set(s, 'format', 'csv')
            sections.append(s)
            
        elif len(args) == 1:
            if config.has_section(args[0]):
                # apply given section parameters, then load from stdin
                config.set(args[0], 'filename', 'sys.stdin')
                sections.append(args[0])
            else:
                print >>sys.stderr, \
                      "Error: Please provide a [%s] section" % args[0]
                sys.exit(5)
        else:
            print >>sys.stderr, \
                  "Error: can't read several sections all from stdin"
            sys.exit(5)

    elif len(args) > 0:
        for s in args:
            if config.has_section(s):
                sections.append(s)
            else:
                log.info("Creating a section for file '%s'" % s)
                # a filename was given, apply [pgsql] defaults
                # set the tablename as the filename sans extension
                # consider columns = *
                if not os.path.exists(s):
                    print >>sys.stderr, \
                        "Error: '%s' does not exists as a section nor as a file" % s
                    sys.exit(2)

                config.add_section(s)
                config.set(s, 'table', os.path.splitext(os.path.basename(s))[0])
                config.set(s, 'filename', s)
                config.set(s, 'columns', '*')
                config.set(s, 'format', 'csv')
                sections.append(s)

    else:
        if not LOAD_FROM_STDIN:
            # don't load all sections first when asked to load stdin
            log.debug("No argument on CLI, will consider all sections")
            for s in config.sections():
                if s != 'pgsql':
                    sections.append(s)

            # we run through sorted section list, unless we got the section list
            # from command line
            sections.sort()

    if FILE_BOUNDARIES is not None and len(sections) > 1:
        print >>sys.stderr, \
              "Error: will not apply boundaries on more than one file"
        sys.exit(5)

    log.info('Will consider following sections:')
    for line in myprint(sections):
        log.info(line)

    # we count time passed from now on
    begin = time.time()

    threads  = {}
    started  = {}
    finished = {}
    current  = 0

    max_running = MAX_PARALLEL_SECTIONS
    if max_running == -1:
        max_running = len(sections)

    log.info('Will load %d section at a time' % max_running)

    sem = threading.BoundedSemaphore(max_running)
    
    while current < len(sections):
        s = sections[current]

        try:
            loader      = None
            summary [s] = []
            started [s] = threading.Event()
            finished[s] = threading.Event()

            try:
                loader = PGLoader(s, config, sem,
                                  (started[s], finished[s]), summary[s])
            except PGLoader_Error, e:
                # could not initialize properly this loader, don't
                # ever wait for it
                started[s] .set()
                finished[s].set()
                log.error(e)
                if DEBUG:
                    raise

            except IOError, e:
                # No space left on device?  can't log it
                break

            if loader:
                if not loader.template:
                    if FILE_BOUNDARIES is not None and len(sections) == 1:
                        loader.reader.set_boundaries(FILE_BOUNDARIES)
                    filename       = loader.filename
                    input_encoding = loader.input_encoding
                    threads[s]     = loader

                    # .start() will sem.aquire(), so we won't have more
                    # than max_running threads running at any time.
                    log.debug("Starting a thread for %s" % s)
                    threads[s].start()
                else:
                    log.info("Skipping section %s, which is a template" % s)

                    for d in (summary, started, finished):
                        d.pop(s)

        except PGLoader_Error, e:
            if e == '':
                log.error('[%s] Please correct previous errors' % s)
            else:
                log.error('%s' % e)

            if DEBUG:
                raise

            if PEDANTIC:
                # was: threads[s].print_stats()
                # but now thread[s] is no more alive
                pass

        except UnicodeDecodeError, e:
            log.error("can't open '%s' with given input encoding '%s'" \
                               % (filename, input_encoding))
                                    
        except IOError, e:
            # typically, No Space Left On Device, can't report nor continue
            break

        current += 1

    # get sure each thread is started, then each one is done
    from pgloader.tools import check_events

    check_events(started, log, "is running")
    log.info("All threads are started, wait for them to terminate")
    check_events(finished, log, "processing is over")

    # check whether any thread failed
    for section, loader in threads.iteritems():
        if not loader.success:
            return 1

    # total duration
    td = time.time() - begin

    if SUMMARY:
        try:
            print_summary(None, sections, summary, td)
            print
        except PGLoader_Error, e:
            log.error("Can't print summary: %s" % e)
            return 1

        except KeyboardInterrupt:
            return 1

    return 0

if __name__ == "__main__":
    # default exit value
    ret = 1

    # Disable KeyboardInterrupt, just exit on CTRL+C
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    try:
        ret = load_data()

    # SystemExit inherits from Exception in python < 2.5, so we have to
    # handle it separately.
    except SystemExit, e:
        sys.exit(e.code)

    except Exception, e:
        from pgloader.options import DEBUG
        if DEBUG:
            raise
        sys.stderr.write(str(e) + '\n')
        sys.exit(1)

    except IOError, e:
        sys.stderr.write(str(e) + '\n')
        sys.exit(1)

    sys.exit(ret)



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader time-related reformating module
#

def time(reject, input):
    """ Reformat str as a PostgreSQL time

    Input time like: 08231560
    We want instead this input: 08:23:15.60
    """
    if len(input) != 8:
        e = "time reformat input too short: %s" % input
        reject.log(e, input)
    
    hour       = input[0:2]
    minute     = input[2:4]
    seconds    = input[4:6]
    hundredths = input[6:8]
    
    return '%s:%s:%s.%s' % (hour, minute, seconds, hundredths)



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader text format reader
#
# handles configuration, parse data, then pass them to database module for
# COPY preparation

import os, sys, os.path, time
from cStringIO import StringIO

from tools    import PGLoader_Error, Reject, parse_config_string
from db       import db
from lo       import ifx_clob, ifx_blob
from reader   import DataReader, UnbufferedFileReader

from options import DRY_RUN, PEDANTIC
from options import TRUNCATE, VACUUM
from options import COUNT, FROM_COUNT, FROM_ID
from options import INPUT_ENCODING, PG_CLIENT_ENCODING
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import NEWLINE_ESCAPES

class TextReader(DataReader):
    """
    Read some text formatted data, which look like CSV but are not:
     - no quoting support
     - trailing separator trailing support
     - multi-line support is explicit (via field_count parameter)
     - newline escaping in multi-line content support
     - ...
    """

    def __init__(self, log, db, reject, filename, input_encoding,
                 table, columns, newline_escapes = None):
        """ init textreader with a newline_escapes parameter """
        DataReader.__init__(self, log, db, reject,
                            filename, input_encoding, table, columns)

        if 'newline_escapes' not in self.__dict__:
            self.newline_escapes = newline_escapes

        self.log.debug('reader.__init__: newline_escapes %s' \
                       % self.newline_escapes)

    def readconfig(self, config, name, template):
        """ get this reader module configuration from config file """
        DataReader.readconfig(self, config, name, template)

        # this will be called twice if templates are in used, so we
        # have to protect ourselves against removing already read
        # configurations while in second run.

        self._getopt('field_count', config, name, template, None, 'int')
        self._getopt('trailing_sep', config, name, template, False)
        if self.trailing_sep is not False:
            self.trailing_sep = self.trailing_sep == 'True'

        self.log.debug('reader.readconfig: field_count %s', self.field_count)
        self.log.debug('reader.readconfig: trailing_sep %s', self.trailing_sep)

    def readlines(self):
        """ read data from configured file, and generate (yields) for
        each data line: line, columns and rowid """

        # temporary feature for controlling when to begin real inserts
        # if first time launch, set to True.
        input_buffer = StringIO()
        nb_lines     = 0
        begin_linenb = None
        nb_plines    = 0

        tmp_offsets  = []
        offsets      = []

        ##
        # if neither -I nor -F was used, we can state that begin = 0
        if FROM_ID is None and self.skip_head_lines == 0:
            self.log.debug('beginning on first line')
            begin_linenb = 1

        self.fd = UnbufferedFileReader(self.filename, self.log,
                                       encoding        = self.input_encoding,
                                       start           = self.start,
                                       end             = self.end,
                                       skip_head_lines = self.skip_head_lines,
                                       check_count     = False)
        
        for line in self.fd:
            # we count real physical lines
            nb_plines += 1

            # and we store them in offsets for error messages
            tmp_offsets.append(nb_plines)
            self.log.debug('current offset %s' % tmp_offsets)

            if self.input_encoding is not None:
                # this may not be necessary, after all
                try:
                    line = line.encode(self.input_encoding)
                except UnicodeDecodeError, e:
                    reject.log(['Codec error', str(e)], input_line)
                    continue
                
            if self.field_count is not None:
                input_buffer.write(line)
                # act as if this were the last input_buffer for this line
                tmp     = self._chomp(input_buffer.getvalue())
                columns = self._split_line(tmp)
                nb_cols = len(columns)

                # check we got them all if not and field_count was
                # given, we have a multi-line input
                if nb_cols < self.field_count:
                    continue
                else:
                    # we have read all the logical line
                    line = tmp
                    input_buffer.close()
                    input_buffer = StringIO()

                    if nb_cols != self.field_count:
                        self.log.debug(line)
                        self.log.debug(str(columns))
                        self.reject.log(
                            'Error parsing columns on line ' +\
                            '%d [row %d]: found %d columns' \
                            % (nb_plines, nb_lines, nb_cols), line)
            else:
                # normal operation mode : one physical line is one
                # logical line. we didn't split input line yet
                line    = self._chomp(line)
                nb_cols = None
                columns = None

            if len(line) == 0:
                # skip empty lines
                continue

            # we count logical lines
            nb_lines  += 1
            offsets    = tmp_offsets[:]
            tmp_offsets = []

            if self.start:
                offsets = (self.start, offsets)

            ##
            # check for beginning if option -I was used
            if FROM_ID is not None:
                if columns is None:
                    columns = self._split_line(line)
                    
                rowids = self._rowids(columns)
                                      
                if FROM_ID == rowids:
                    begin_linenb = nb_lines
                    self.log.debug('reached beginning on line %d', nb_lines)

                elif begin_linenb is None:
                    # begin is set to 1 when we don't use neither -I nor -F
                    continue

            if COUNT is not None and self.fd.reading \
               and (nb_lines - begin_linenb + 1) > COUNT:
                
                self.log.info('reached line %d, stopping', nb_lines)
                break

            if columns is None:
                columns = self._split_line(line)

            self.log.debug('read data')

            # now, we may have to apply newline_escapes on configured columns
            if NEWLINE_ESCAPES or self.newline_escapes != []:
                columns = self._escape_newlines(columns)

            nb_cols = len(columns)
            if nb_cols != len(self.columns):
                self.log.debug(line)
                self.log.debug(str(columns))

                msg = 'Error parsing columns on line ' +\
                      '%d [row %d]: found %d columns' \
                      % (nb_plines, nb_lines, nb_cols)

                self.reject.log(msg, line)
                continue

            yield offsets, line, columns

    def _split_line(self, line):
        """ split given line and returns a columns list """
        last_sep = 0
        columns  = []
        pos      = line.find(self.field_sep, last_sep)
        
        while pos != -1:
            # don't consider backslash escaped separators melted into data
            # warning, we may find some data|\\|data
            # that is some escaped \ then a legal | separator
            i=1
            while pos-i >= 0 and line[pos-i] == '\\':
                i += 1

            # now i-1 is the number of \ preceding the separator
            # it's a legal separator only if i-1 is even
            if (i-1) % 2 == 0:
                # there's no need to keep escaped delimiters inside a column
                # and we want to avoid double-escaping them
                columns.append(line[last_sep:pos]
                               .replace("\\%s" % self.field_sep,
                                        self.field_sep))
                last_sep = pos + 1

            pos = line.find(self.field_sep, pos + 1)

        # append last column
        columns.append(line[last_sep:])
        return columns


    def _chomp(self, input_line):
        """ chomp end of line when necessary, and trailing_sep too """

        if len(input_line) == 0:
            self.log.debug('pgloader._chomp: skipping empty line')
            return input_line
        
        # chomp a copy of the input_line, we will need the original one
        line = input_line[:]

        if line[-2:] == "\r\n":
            line = line[:-2]

        elif line[-1] == "\r":
            line = line[:-1]

        elif line[-1] == "\n":
            line = line[:-1]

        # trailing separator to whipe out ?
        if self.trailing_sep \
            and line[-len(self.field_sep)] == self.field_sep:
            
            line = line[:-len(self.field_sep)]

        return line

    def _escape_newlines(self, columns):
        """ trim out newline escapes to be found inside data columns """
        self.log.debug('escaping columns newlines')
        self.log.debug(self.newline_escapes)

        for (ne_col, ne_esc) in self.newline_escapes:
            # don't forget configured col references use counting from 1
            ne_colnum = dict(self.columns)[ne_col] - 1
            self.log.debug('column %s[%d] escaped with %s',
                          ne_col, ne_colnum+1, ne_esc)

            col_data = columns[ne_colnum]
            
            if self.db and \
               (self.db.is_null(col_data) or self.db.is_empty(col_data)):
                self.log.debug('skipping null or empty column')
                continue

            escaped   = []
            tmp       = col_data

            for line in tmp.split('\n'):
                if len(line) == 0:
                    self.log.debug('skipping empty line')
                    continue

                self.log.debug('chomping: %s', line)
                    
                tmpline = self._chomp(line)
                if tmpline[-1] == ne_esc:
                    tmpline = tmpline[:-1]

                    # chomp out only escaping char, not newline itself
                    escaped.append(line[:len(tmpline)] + \
                                   line[len(tmpline)+1:])

                else:
                    # line does not end with escaping char, keep it
                    escaped.append(line)

            columns[ne_colnum] = '\n'.join(escaped)
        return columns




# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader data reader interface and defaults

import sys
from tools    import PGLoader_Error, Reject, parse_config_string
from db       import db
from lo       import ifx_clob, ifx_blob

from options import DRY_RUN, PEDANTIC
from options import TRUNCATE, VACUUM
from options import COUNT, FROM_COUNT, FROM_ID
from options import INPUT_ENCODING, PG_CLIENT_ENCODING
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import NEWLINE_ESCAPES

class DataReader:
    """
    Read some text formatted data, which look like CSV but are not:
     - no quoting support
     - multi-line support is explicit (via 
    """

    def __init__(self, log, db, reject,
                 filename, input_encoding, table, columns):
        """ init internal variables """
        log.debug('reader __init__ %s %s %s', filename, table, columns)

        self.log       = log
        self.db        = db
        self.filename  = filename
        self.input_encoding = input_encoding
        self.table     = table
        self.columns   = columns
        self.reject    = reject
        self.mem_units = {'kB': 1024,
                          'MB': 1024*1024,
                          'GB': 1024*1024*1024,
                          'TB': 1024*1024*1024*1024}

        if self.input_encoding is None:
            if INPUT_ENCODING is not None:
                self.input_encoding = INPUT_ENCODING

        # (start, end) are used for split_file_reading mode
        # queue when in round_robin_read mode
        self.start = None
        self.end   = None

    def readconfig(self, config, name, template):
        """ read configuration section for common options

        name is configuration section name, conf the ConfigParser object

        template is the (maybe None) template section name declared in
        the use_template configuration option.

        specific option reading code is to be found on subclasses
        which implements read data parsing code.

        see textreader.py and csvreader.py
        """

        if not DRY_RUN:
            # optionnal null and empty_string per table parameters
            self._getopt('null', config, name, template, NULL)
            self.db.null = parse_config_string(self.null)

            self._getopt('empty_string', config, name, template, EMPTY_STRING)
            self.db.empty_string = parse_config_string(self.empty_string)

        self._getopt('field_sep', config, name, template, FIELD_SEP)
        self.field_sep = self.field_sep.decode('string-escape')

        ##
        # FROM_COUNT takes precedence over skip_head_lines
        if FROM_COUNT is None or FROM_COUNT == 0:
            self._getopt('skip_head_lines', config, name, template, 0, 'int')
        else:
            self.skip_head_lines = FROM_COUNT - 1

        if len(self.field_sep) != 1:
            raise PGLoader_Error, "field_sep must be 1 char, not %d (%s)" \
                  % (len(self.field_sep), self.field_sep)
        
        if not DRY_RUN:
            if self.db.copy_sep is None:
                self.db.copy_sep = self.field_sep

        if not DRY_RUN:
            self.log.debug("reader.readconfig null: '%s'" % self.db.null)
            self.log.debug("reader.readconfig empty_string: '%s'",
                           self.db.empty_string)
            self.log.debug("reader.db %s copy_sep %s" % (self.db, self.db.copy_sep))
            
        self.log.debug("reader.readconfig field_sep: '%s'", self.field_sep)
        self.log.debug("reader.readconfig skip_head_lines: %d",
                       self.skip_head_lines)

    def _getopt(self, option, config, section, template, default = None, opt_type = "char"):
        """ Init given configuration option """

        if config.has_option(section, option):
            self.__dict__[option] = config.get(section, option)
            self.log.debug("reader._getopt %s from %s is '%s'" % (option, section, self.__dict__[option]))

        elif template and config.has_option(template, option):
            self.__dict__[option] = config.get(template, option)
            self.log.debug("reader._getopt %s from %s is '%s'" \
                           % (option, template, self.__dict__[option]))

        elif option not in self.__dict__:
            self.log.debug("reader._getopt %s defaults to '%s'" \
                           % (option, default))
            self.__dict__[option] = default

        if opt_type == 'int' and self.__dict__[option] is not None:
            try:
                self.__dict__[option] = int(self.__dict__[option])
            except ValueError, e:
                self.log.error('Configuration option %s.%s is not an int: %s' \
                               % (section, option, self.__dict__[option]))
                raise PGLoader_Error, e

        elif opt_type == 'mem' and self.__dict__[option] is not None:
            try:
                opt = self.__dict__[option]
                if type(opt) == type("string") \
                       and len(opt) > 2 and opt [-2:] in self.mem_units:
                    unit = opt[-2:]
                    size = int(opt[:-2]) * self.mem_units[unit]
                    self.__dict__[option] = int(size)
                else:
                    self.__dict__[option] = int(self.__dict__[option])
            except ValueError, e:
                self.log.error('Configuration option %s.%s is not a memsize: %s' \
                               % (section, option, self.__dict__[option]))
                raise PGLoader_Error, e

        return self.__dict__[option]

    def readlines(self):
        """ read data from configured file, and generate (yields) for
        each data line: offsets, line, columns and rowid 

        offsets: list of line numbers where the line was read in the file
                 tuple of (reader offset, [list, of, line, numbers])

        the second case is used when in split file reading mode
        """
        pass

    def set_boundaries(self, (start, end)):
        """ set the boundaries of this reader """
        self.start = start
        self.end   = end

        self.log.debug("reader start=%d, end=%d" % (self.start, self.end))


class UnbufferedFileReader:
    """
    Allow to read a file line by line, avoiding any read-buffering
    effect. This allows for readers to reliably compare fd.tell()
    position after each line reading.
    """

    def __init__(self, filename, log,
                 mode = "rb", encoding = None,
                 start = None, end = None,
                 skip_head_lines = 0,
                 check_count = True):
        """ constructor """
        self.filename = filename
        self.log      = log
        self.mode     = mode
        self.encoding = encoding
        self.start    = start
        self.end      = end
        self.fd       = None
        self.position = 0
        self.line_nb  = 0

        # check_count can be set to False when phisical lines and logical
        # lines counts can diverge, like in textreader.py
        self.check_count = check_count
        self.skip_head_lines = skip_head_lines
        self.reading = self.skip_head_lines == 0

        # we don't yet force buffering, but...
        self.bufsize = -1
        
        if self.encoding is not None:
            try:
                import codecs
                if self.filename == 'sys.stdin':
                    f = sys.stdin
                else:
                    f = open(self.filename, self.mode, self.bufsize)

                self.fd = codecs.getreader(self.encoding)(f)
                self.log.info("Opened '%s' with encoding '%s'" \
                              % (self.filename, self.encoding))
            except LookupError, e:
                # codec not found
                raise PGLoader_Error, "Input codec: %s" % e
            except IOError, e:
                # file not found, for example
                raise PGLoader_Error, "IOError: %s" % e
            
        else:
            try:
                if self.filename == 'sys.stdin':
                    self.fd = sys.stdin
                else:
                    self.fd = open(self.filename, mode, self.bufsize)
            except IOError, error:
                raise PGLoader_Error, error

        if self.start:
            self.fd.seek(self.start)
            self.position = self.fd.tell()

        self.log.info("Opened '%s' in %s (fileno %s), ftell %d" \
                      % (self.filename, self.fd,
                         self.fd.fileno(), self.position))
        return

    def tell(self):
        return self.position

    def seek(self, position):
        self.fd.seek(position)
        self.position = self.fd.tell()

        return self.position
        
    def next(self):
        """ implement the iterator protocol """
        yield self.__iter__()

    def __iter__(self):
        """ read a line at a time """
        line = 'dumb non-empty init value'
        last_line_read = False
        
        while line != '':
            line = self.fd.readline()
            self.line_nb += 1

            ## try:
            ##     self.position = self.fd.tell()
            ## except IOError, error:
            ##     #IOError: [Errno 29] Illegal seek --- when stdin reaches EOF
            ##     self.log.info(error)
            ##     return

            ##
            # if -F is used, count lines to skip, and skip them
            if self.skip_head_lines > 0:
                if self.line_nb <= self.skip_head_lines:
                    continue

                if self.line_nb == self.skip_head_lines + 1:
                    self.reading = True
                    self.log.info('reached beginning on line %d', self.line_nb)


            # check if we already processed COUNT lines
            if self.check_count:
                if COUNT is not None and self.reading \
                   and (self.line_nb - self.skip_head_lines + 1) > COUNT:

                    self.log.info('reached line %d, stopping', self.line_nb)
                    return

            # check EOF (real or multi-readers)
            if line == '' or last_line_read:
                try:
                    self.log.debug("FileReader stoping, offset %d >= %s" \
                                   % (self.fd.tell(), self.end))
                except IOError, error:
                    #IOError: [Errno 29] Illegal seek --- when stdin reaches
                    # EOF should not happen as --load-from-stdin and
                    # --boundaries are not accepted at the same time
                    self.log.info(error)
                
                self.fd.close()
                return

            # check multi-reader boundaries
            if self.end is not None and self.fd.tell() >= self.end:
                # we want to process current line and stop at next
                # iteration
                self.log.info("Reached position %d, reading last line" \
                              % self.fd.tell())
                last_line_read = True

            if self.encoding is not None:
                yield line.encode(self.encoding)
            else:
                yield line

        return
    



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# Some common options, for each module to get them

from tempfile import gettempdir
import os

PGLOADER_VERSION = '2.3.3~dev3'

PSYCOPG_VERSION = None

INPUT_ENCODING     = None
PG_CLIENT_ENCODING = 'latin9'
PG_OPTIONS         = None
DATESTYLE          = None

COPY_SEP     = None
FIELD_SEP    = '|'
CLOB_SEP     = ','
NULL         = ''
EMPTY_STRING = '\ '

NEWLINE_ESCAPES = None

DEBUG      = False
VERBOSE    = False
QUIET      = False
SUMMARY    = False
DRY_RUN    = False
PEDANTIC   = False

TRUNCATE   = False
VACUUM     = False
TRIGGERS   = False

COUNT      = None
FROM_COUNT = None
FROM_ID    = None

UDC_PREFIX = 'udc_'

REFORMAT_PATH = None
DEFAULT_REFORMAT_PATH = ['/usr/share/python-support/pgloader/reformat']

DEFAULT_MAX_PARALLEL_SECTIONS = 1
DEFAULT_SECTION_THREADS       = 1
MAX_PARALLEL_SECTIONS = None
SECTION_THREADS       = None
SPLIT_FILE_READING    = False
RRQUEUE_SIZE          = None

CLIENT_MIN_MESSAGES = None
LOG_MIN_MESSAGES    = DEBUG
DEFAULT_LOG_FILE    = os.path.join (gettempdir(), "pgloader.log")
LOG_FILE            = None

REJECT_LOG_FILE  = '%s.rej.log'
REJECT_DATA_FILE = '%s.rej'

LOAD_FROM_STDIN = None
LOAD_TO_TABLE   = None

FILE_BOUNDARIES = None  # (start, end) --- file positions in bytes



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader mysql reformating module
#

def timestamp(reject, input):
    """ Reformat str as a PostgreSQL timestamp

    MySQL timestamps are like:  20041002152952
    We want instead this input: 2004-10-02 15:29:52
    """
    if len(input) != 14:
        e = "MySQL timestamp reformat input too short: %s" % input
        reject.log(e, input)
    
    year    = input[0:4]
    month   = input[4:6]
    day     = input[6:8]
    hour    = input[8:10]
    minute  = input[10:12]
    seconds = input[12:14]
    
    return '%s-%s-%s %s:%s:%s' % (year, month, day, hour, minute, seconds)



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader main class
#
# handles configuration, parse data, then pass them to database module for
# COPY preparation

import os, sys, os.path, time, codecs, threading
from cStringIO import StringIO
from tempfile import gettempdir

from logger   import log, getLogger
from tools    import PGLoader_Error, Reject, parse_config_string, check_events
from db       import db
from lo       import ifx_clob, ifx_blob

from options import DRY_RUN, PEDANTIC
from options import TRUNCATE, VACUUM, TRIGGERS
from options import COUNT, FROM_COUNT, FROM_ID
from options import INPUT_ENCODING, PG_CLIENT_ENCODING
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import PG_OPTIONS
from options import NEWLINE_ESCAPES
from options import UDC_PREFIX
from options import REFORMAT_PATH
from options import REJECT_LOG_FILE, REJECT_DATA_FILE
from options import MAX_PARALLEL_SECTIONS
from options import DEFAULT_SECTION_THREADS, SECTION_THREADS, SPLIT_FILE_READING
from options import RRQUEUE_SIZE

class PGLoader(threading.Thread):
    """
    PGLoader reads some data file and depending on ts configuration,
    import data with COPY or update blob data with UPDATE.
    """

    def __init__(self, name, config, sem, (started, finished),
                 stats,
                 logname = None,
                 reject = None, queue = None, lock = None, copy_sep = None):
        """ Init with a configuration section """
        
        # logname is given when we use several Threads for reading
        # input file
        self.logname = logname
        if self.logname is None:
            self.logname = name
        self.log = getLogger(self.logname)

        threading.Thread.__init__(self, name = self.logname)

        # sem, stats and queue (if not None) are global objects
        # sem       is shared by all threads at the same level
        # started   is a threads.Event object to set() once in run()
        # finished  is a threads.Event object to set() once processing is over
        # stats     is a private entry of a shared dict
        # queue     is given when reading in round robin mode
        # lock      is a threading.Lock for reading sync
        # reject    is the common reject object
        #
        self.sem      = sem
        self.started  = started
        self.stats    = stats
        self.finished = finished
        self.queue    = queue
        self.lock     = lock
        self.reject   = reject

        # thereafter parameters are local
        self.name    = name
        self.config  = config

        # in the round-robin reader case, workers won't have
        # initialized a reader, thus won't read the configuration of
        # copy_sep.
        self.copy_sep = copy_sep

        self.template     = None
        self.use_template = None
        self.tsection     = None

        self.index     = None
        self.all_cols  = None
        self.columns   = None
        self.blob_cols = None

        self.config_errors = 0
        self.errors        = 0
        self.updates       = 0
        self.init_time     = None

        # we may have to open several clob files while parsing the
        # unload data file, hence we keep track of them all
        self.blobs = {}

        if config.has_option(name, 'template'):
            self.template = True
            self.log.debug("[%s] is a template", self.name)

        if not self.template:
            self.log.debug("[%s] parse configuration", self.name)

        if not self.template:
            # check if the section wants to use a template
            if config.has_option(name, 'use_template'):
                self.tsection = config.get(name, 'use_template')

        self._dbconfig(config)

        if self.tsection is not None:
            self.template = config.get(name, 'use_template')

            if not config.has_section(self.template):
                m = '%s refers to unknown template section %s' \
                    % (name, self.template)

                raise PGLoader_Error, m

            # first load template configuration
            self.log.debug("Reading configuration from template " +\
                           "section [%s]", self.template)

            self.real_log = self.log
            self.log = getLogger(self.template)

            try:
                self._read_conf(self.template, config, db,
                                want_template = True)
            except PGLoader_Error, e:
                self.log.error(e)
                m = "%s.use_template does not refer to a template section"\
                    % name
                raise PGLoader_Error, m

            # reinit self.template now its relative config section is read
            self.template = None
            self.log      = self.real_log

        if not self.template:
            # now load specific configuration
            self.log.debug("Reading configuration from section [%s]", name)
            self._read_conf(name, config, db)

        self.log.debug('%s init done' % name)

    def __del__(self):
        """ PGLoader destructor, we close the db connection """
        if not DRY_RUN:
            self.db.close()

    def _dbconfig(self, config):
        """ connects to database """
        section = 'pgsql'

        if DRY_RUN:
            log.info("dry run mode, not connecting to database")
            self.db = None
            return

        try:
            dsn = []
            if config.has_option(section, 'host'):
                dsn.append ("host=%s" % config.get(section, 'host'))
            if config.has_option(section, 'port'):
                dsn.append ("port=%d" % config.getint(section, 'port'))
            if config.has_option(section, 'base'):
                dsn.append ("dbname=%s" % config.get(section, 'base'))
            if config.has_option(section, 'user'):
                dsn.append ("user=%s" % config.get(section, 'user'))
            if config.has_option(section, 'pass'):
                dsn.append ("password=%s" % config.get(section, 'pass'))
            if config.has_option(section, 'sslmode'):
                dsn.append ("sslmode=%s" % config.get(section, 'sslmode'))

            self.db = db(" ".join (dsn), connect = False)

            for opt in ['client_encoding', 'datestyle', 'lc_messages']:
                if config.has_option(section, opt):
                    self.db.pg_options[opt] = \
                        parse_config_string(config.get(section, opt))

            # PostgreSQL options
            from tools import parse_pg_options
            parse_pg_options(self.log, config, section, self.db.pg_options)
            self.log.debug("_dbconfig: %s" % str(self.db.pg_options))

            if config.has_option(section, 'copy_every'):
                self.db.copy_every = config.getint(section, 'copy_every')

            if config.has_option(section, 'commit_every'):
                self.db.commit_every = config.getint(
                    section, 'commit_every')

            if self.copy_sep is not None:
                self.db.copy_sep = self.copy_sep
                self.log.debug("got copy_sep '%s' from self.copy_sep" \
                               % self.copy_sep)

            if config.has_option(section, 'copy_delimiter'):
                self.db.copy_sep = config.get(section, 'copy_delimiter')
                self.log.debug("got copy_sep '%s' from %s" \
                               % (self.db.copy_sep, section))

            elif config.has_option(self.tsection, 'copy_delimiter'):
                self.db.copy_sep = config.get(section, 'copy_delimiter')
                self.log.debug("got copy_sep '%s' from self.copy_sep" \
                               % (self.db.copy_sep, self.tsection))

            self.log.debug("_dbconnect copy_sep %s " % self.db.copy_sep)

            # we want self.db messages to get printed as from our section
            self.db.log = self.log

        except Exception, error:
            log.error("Could not initialize PostgreSQL connection")
            raise PGLoader_Error, error

    def _postinit(self):
        """ This has to be called while self.sem is acquired """
        # Now reset database connection
        if not DRY_RUN:
            self.db.reset()

        if not self.template and not DRY_RUN:
            # check we have properly configured the copy separator
            if self.db.copy_sep is None:
                self.log.debug("%s" % self.db)
                self.log.error("COPY sep is %s" % self.db.copy_sep)
                msg = "BUG: pgloader couldn't configure its COPY separator"
                raise PGLoader_Error, msg
        
    def _read_conf(self, name, config, db, want_template = False):
        """ init self from config section name  """

        # we'll need both of them from the globals
        global FROM_COUNT, FROM_ID

        if want_template and not config.has_option(name, 'template'):
            e = 'Error: section %s is not a template' % name
            raise PGLoader_Error, e

        ##
        # reject log and data files defaults to /tmp/<section>.rej[.log]
        if self.reject is None:
            # If we've been given a reject object, just use it, don't
            # even try to init a new one from configuration
            
            if config.has_option(name, 'reject_log'):
                self.reject_log = config.get(name, 'reject_log')

            if config.has_option(name, 'reject_data'):
                self.reject_data = config.get(name, 'reject_data')

            if not self.template and 'reject_log' not in self.__dict__:
                self.reject_log = os.path.join(gettempdir(), REJECT_LOG_FILE % name)

            if not self.template and 'reject_data' not in self.__dict__:
                self.reject_data = os.path.join(gettempdir(), REJECT_DATA_FILE % name)

            # reject logging
            if not self.template:
                self.reject = Reject(self.log,
                                     self.reject_log, self.reject_data)

                self.log.debug('reject log in %s', self.reject.reject_log)
                self.log.debug('rejected data in %s', self.reject.reject_data)

            else:
                # needed to instanciate self.reject while in template section
                self.reject = None

        # optionnal local option input_encoding
        self.input_encoding = INPUT_ENCODING
        if config.has_option(name, 'input_encoding'):
            self.input_encoding = parse_config_string(
                config.get(name, 'input_encoding'))
        self.log.debug("input_encoding: '%s'", self.input_encoding)

        # optionnal local option client_encoding and datestyle
        for opt in ['client_encoding', 'datestyle']:
            if config.has_option(name, opt):
                self.db.pg_options[opt] = parse_config_string(config.get(name, opt))

                if not DRY_RUN:
                    self.log.debug("%s: '%s'", opt, self.db.pg_options[opt])

        # optionnal local pg_options
        # precedence is given to command line parsing, which is in PG_OPTIONS
        from tools import parse_pg_options
        parse_pg_options(log, config, name, self.db.pg_options, overwrite=True)
        if not self.template:
            if PG_OPTIONS:
                self.db.pg_options.update(PG_OPTIONS)

        ##
        # data filename
        for opt in ('table', 'filename'):
            if config.has_option(name, opt):
                self.log.debug('%s.%s: %s', name, opt, config.get(name, opt))
                self.__dict__[opt] = config.get(name, opt)
            else:
                if not self.template and opt not in self.__dict__:
                    msg = "Error: Please configure %s.%s" % (name, opt)
                    raise PGLoader_Error, msg
                
                elif not self.template and not self.__dict__[opt]:
                    self.log.error('Error: please configure %s.%s', name, opt)
                    self.config_errors += 1
                    
                else:
                    # Reading Configuration Template section
                    # we want the attribute to exists for further usage
                    if opt not in self.__dict__:
                        self.__dict__[opt] = None

        ##
        # we parse some columns definitions
        if config.has_option(name, 'index'):
            self._parse_fields('index', config.get(name, 'index'))

        if config.has_option(name, 'columns'):
            conf_cols = config.get(name, 'columns')
            if conf_cols == '*':
                self.all_cols    = True
                self.db.all_cols = True

                # get column list from database
                self.columns = self.db.get_all_columns(self.table)
                self.log.info("columns = *, got %s", str(self.columns))

            else:
                self._parse_fields('columns', config.get(name, 'columns'))

        if config.has_option(name, 'blob_columns'):
            self._parse_fields('blob_cols',
                               config.get(name, 'blob_columns'),
                               btype = True)

        self.log.debug('index %s', str(self.index))
        self.log.debug('columns %s', str(self.columns))
        self.log.debug('blob_columns %s', str(self.blob_cols))

        if self.columns is None:
            if not self.template:
                self.log.error('%s has no columns defined', name)
                self.config_errors += 1

            else:
                # non critical error, and code thereafter wants to use
                # self.columns as a list
                self.columns = []

        ##
        # The config section can also provide user-defined colums
        # which are option beginning with options.UDC_PREFIX
        udcs = [o
                for o in config.options(name)
                if o[:len(UDC_PREFIX)] == UDC_PREFIX]
        
        if len(udcs) > 0:
            self.udcs = []
            for udc in udcs:
                udc_name  = udc[:]
                udc_name  = udc_name[udc_name.find('_')+1:]
                udc_value = config.get(name, udc)

                self.udcs.append((udc_name, udc_value))
        else:
            self.udcs = None

        self.log.debug('udcs: %s', str(self.udcs))

        # better check there's no user defined column overriding file
        # columns
        if self.udcs:
            errs = []
            cols = [c for (c, cn) in self.columns]
            for (udc_name, udc_value) in self.udcs:
                if udc_name in cols:
                    errs.append(udc_name)

            if errs:
                for c in errs:
                    self.log.error('%s is configured both as a %s.columns ' +\
                                   'entry and as a user-defined column',
                                   c, name)

                self.config_errors += 1

        # we need the copy_columns parameter if user-defined columns
        # are used
        #
        # when using templates, we can read copy_columns setup without
        # having any user-defined column defined
        if self.template or self.udcs:
            if config.has_option(name, 'copy_columns'):
                namelist = [n for (n, c) in self.columns]
                if self.udcs:
                    namelist += [n for (n, v) in self.udcs]

                self.config_copy_columns = config.get(name, 'copy_columns')
                self.config_copy_columns = self.config_copy_columns.split(',')
                
                self.copy_columns = []
                for x in self.config_copy_columns:
                    x = x.strip(' \n\r')
                    # just add all the given columns in this pass
                    self.copy_columns.append(x)

                self.log.debug('config copy_columns %s',
                               str(self.config_copy_columns))
                self.log.debug('copy_columns %s', str(self.copy_columns))
                    
        if not self.template:
            # check for errors time!
            if 'copy_columns' in self.__dict__:
                namelist = [n for (n, c) in self.columns] + \
                           [n for (n, v) in self.udcs]

                for x in self.copy_columns:
                    if x not in namelist:
                        self.log.error('"%s" not in %s column list, ' + \
                                       'including user defined columns',
                                       x, name)
                        self.config_errors += 1
                        
                if len(self.copy_columns) != len(self.config_copy_columns):
                    self.log.error('%s.copy_columns refers to ' +\
                                   'unconfigured columns', name)

                    self.config_errors += 1
                    
            elif self.udcs:
                self.log.error('section %s does not define copy_columns ' +\
                               'but uses user-defined columns', name)

                self.config_errors += 1

        # in the copy_columns case, columnlist is that simple:
        self.columnlist = None
        if self.udcs:
            if 'copy_columns' in self.__dict__ and self.copy_columns:
                self.columnlist = self.copy_columns

        self.log.debug('udcs: %s', str(self.udcs))
        if self.udcs and 'copy_columns' in self.__dict__:
            self.log.debug('copy_columns %s', str(self.copy_columns))

        ##
        # We have for example columns = col1:2, col2:1
        # this means the order of input columns is not the same as the
        # awaited order of COPY, so we want a mapping index, here [2, 1]
        #
        # The column mapping is to be done on all_columns, which
        # allows user to have their user-defined columns talken into
        # account in the COPY ordering.
        self.col_mapping = [i for (c, i) in self.columns]

        if self.col_mapping == range(1, len(self.columns)+1):
            # no mapping to do
            self.col_mapping = None

        ##
        # optionnal partial loading option (sequences case)
        #
        # self.columnlist is the column list to give to
        # COPY table(...) command, either the cols given in
        # the only_cols config, or the columns directly
        
        self.only_cols = None

        if config.has_option(name, 'only_cols'):
            if self.udcs:
                self.log.error('section %s defines both ' +\
                               'user-defined columns and only_cols', name)

                self.config_errors += 1
            
            self.only_cols = config.get(name, 'only_cols')

            ##
            # first make an index list out of configuration
            # which contains coma separated ranges or values
            # as for example: only_cols = 1-3, 5
            try:
                only_cols = [x.strip() for x in self.only_cols.split(",")]
                expanded  = []

                # expand ranges
                for oc in only_cols:
                    if '-' in oc:
                        (a, b) = [int(x) for x in oc.split("-")]
                        for i in range(a, b+1):
                            expanded.append(i)
                    else:
                        expanded.append(int(oc))

                # we have to find colspec based on self.columns
                self.only_cols  = expanded
                self.columnlist = [self.columns[x-1][0] for x in expanded]

            except Exception, e:
                self.log.error('section %s, only_cols: ' +\
                               'configured range is invalid', name)
                raise PGLoader_Error, e

        if self.only_cols is None:
            if self.columnlist is None:
                # default case, no user-defined cols, no restriction
                self.columnlist = [n for (n, pos) in self.columns]

        self.log.debug("only_cols %s", str(self.only_cols))
        self.log.debug("columnlist %s", str(self.columnlist))

        if not self.template and self.only_cols is not None:
            self.db.all_cols = False

        ##
        # This option is textreader specific, but being lazy and
        # short-timed, I don't make self._parse_fields() callable from
        # outside this class. Hence the code here.
        #
        # optionnal newline escaped option
        self.newline_escapes = []
        if config.has_option(name, 'newline_escapes'):
            if NEWLINE_ESCAPES is not None:
                # this parameter is globally set, will ignore local
                # definition
                self.log.warning("ignoring %s newline_escapes option" +\
                                 "option is set to '%s' globally",
                                 name, NEWLINE_ESCAPES)
            else:
                self._parse_fields('newline_escapes',
                                   config.get(name, 'newline_escapes'),
                                   argtype = 'char')

        if NEWLINE_ESCAPES is not None:
            # set NEWLINE_ESCAPES for each table column
            self.newline_escapes = [(a, NEWLINE_ESCAPES)
                                    for (a, x) in self.columns]

        self.log.debug("self.newline_escapes = '%s'" % self.newline_escapes)

        ##
        # Parallelism knobs, give preference to command line
        if SECTION_THREADS:
            self.section_threads = SECTION_THREADS
        elif config.has_option(name, 'section_threads'):
            self.section_threads = config.getint(name, 'section_threads')
        else:
            self.section_threads = DEFAULT_SECTION_THREADS

        if not self.template:
            # only log the definitive information
            self.log.info("Loading threads: %d" % self.section_threads)

        if config.has_option(name, 'split_file_reading'):
            self.split_file_reading = config.get(name, 'split_file_reading') == 'True'
        else:
            self.split_file_reading = SPLIT_FILE_READING

        self.rrqueue_size = RRQUEUE_SIZE
        if config.has_option(name, 'rrqueue_size'):
            self.rrqueue_size = config.getint(name, 'rrqueue_size')

        if self.rrqueue_size is None or self.rrqueue_size < 1:
            if DRY_RUN:
                # won't be used
                self.rrqueue_size = 1
            else:
                self.rrqueue_size = self.db.copy_every
            
        if not self.template:
            for opt in ('section_threads', 'split_file_reading'):
                self.log.debug('%s.%s = %s' % (name, opt, str(self.__dict__[opt])))

        if not self.template and self.split_file_reading:
            opt = 'split_file_reading'
            if FROM_COUNT is not None and FROM_COUNT > 0:
                raise PGLoader_Error, \
                      "Conflict: can't use both '%s' and '--from'" % opt

            if FROM_ID is not None:
                raise PGLoader_Error, \
                      "Conflict: can't use both '%s' and '--from-id'" % opt

        ##
        # Reader's init
        if config.has_option(name, 'format'):
            self.format = config.get(name, 'format')
        
        if not self.template and self.queue is None:
            # Only init self.reader in real section, not from
            # template.  self.reader.readconfig() will care about
            # reading its configuration from template and current
            # section.
            #
            # If self.queue is not None, we'll read data from it and
            # not from self.reader, so we don't care about it.

            if 'format' not in self.__dict__:
                raise PGLoader_Error, "Please configure %s.format" % name
                
            self.log.debug("File '%s' will be read in %s format" \
                           % (self.filename, self.format))

            if self.format.lower() == 'csv':
                from csvreader import CSVReader 
                self.reader = CSVReader(self.log, self.db, self.reject,
                                        self.filename,
                                        self.input_encoding,
                                        self.table, self.columns)

            elif self.format.lower() == 'text':
                from textreader import TextReader
                self.reader = TextReader(self.log, self.db, self.reject,
                                         self.filename,
                                         self.input_encoding,
                                         self.table, self.columns,
                                         self.newline_escapes)
                
            elif self.format.lower() == 'fixed':
                from fixedreader import FixedReader
                self.reader = FixedReader(self.log, self.db, self.reject,
                                          self.filename,
                                          self.input_encoding,
                                          self.table, self.columns)
                
            else:
                self.log.error("unknown format '%s'")
                raise PGLoader_Error, "Skipping section %s" % self.name

            self.log.debug('reader.readconfig()')
            self.reader.readconfig(config, name, self.tsection)


            if self.split_file_reading:
                if self.format.lower() == 'text' \
                   and (self.reader.field_count is not None \
                        or self.reader.trailing_sep):

                    # split_file_reading is not compatible with text
                    # mode when field_count or trailing_sep is used
                    # (meaning end of line could be found in the data)

                    raise PGLoader_Error, \
                          "Can't use split_file_reading with text " +\
                          "format when 'field_count' or 'trailing_sep' is used"

        ##
        # Some column might need reformating
        if config.has_option(name, 'reformat'):
            self._parse_fields('c_reformat', config.get(name, 'reformat'),
                               btype = True, argtype = 'string')
        else:
            # remember reformat could have been configured from template
            if 'c_reformat' not in self.__dict__:
                self.c_reformat = self.reformat = None

        self.log.debug('reformat %s', str(self.c_reformat))

        # check the configure reformating is available
        if not self.template and self.c_reformat:
            import imp
            self.reformat = []
            
            for r_colname, r_module, r_function in self.c_reformat:
                if r_colname not in self.columnlist:
                    self.log.error('%s.reformat refers to unknown column %s',
                                   name, r_colname)
                    self.config_errors += 1

                # load the given module name and function
                module = None
                try:
                    fp, pathname, description = \
                        imp.find_module(r_module, REFORMAT_PATH)

                    self.log.debug('Found %s at %s', r_module, str(pathname))
                    
                    module = imp.load_module(r_module,
                                             fp, pathname, description)
                    
                except ImportError, e:
                    self.log.error('%s failed to import reformat module "%s"',
                                   name, r_module)
                    self.log.error('from %s', str(REFORMAT_PATH))
                    self.log.error(e)
                    self.config_errors += 1


                if module:
                    if r_function in module.__dict__:
                        self.reformat.append((r_colname,
                                              module.__dict__[r_function]))
                    else:
                        self.log.error('reformat module %s has no %s function',
                                       r_module, r_function)
                        self.config_errors += 1

        if not self.template:
            self.log.debug('reformat %s', str(self.reformat))

        ##
        # How can we mix those columns definitions ?
        #  - we want to load table data with COPY
        #  - we want to load blob data into text or bytea field, with COPY
        #  - we want to do both
        
        if self.columns is None and self.blob_cols is None:
            # nothing to do ? perfect, done
            self.reject.log(
                "Error: in section '%s': " % self.name +\
                "please configure some work to do (columns, blob_cols)")
            self.config_errors += 1

        if self.blob_cols is not None and self.index is None:
            # if you want to load blobs, UPDATE need indexes
            # is this error still necessary?
            self.reject.log(
                "Error: in section '%s': " % self.name +\
                "for blob importing (blob_cols), please configure index")
            self.config_errors += 1


        ##
        # if options.fromid is not None it has to be either a value,
        # when index is single key or a dict in a string, when index
        # is a multiple key
        if FROM_ID is not None:
            if len(self.index) > 1:
                # we have to evaluate given string and see if it is a
                # dictionnary
                try:
                    d = {}
                    ids  = [x.strip() for x in FROM_ID.split(',')]
                    for id in ids:
                        k, v = id.split(':')
                        d[k] = v

                    FROM_ID = d
                except Exception, e:
                    self.reject.log(
                        'Error: unable to parse given key %s' % FROM_ID)
                    raise PGLoader_Error

            self.log.info('composite key found, -I evaluated to %s', FROM_ID)

        if self.config_errors > 0:
            mesg = 'Configuration errors for section %s' % self.name
            raise PGLoader_Error, mesg

    def _parse_fields(self, attr, str, btype = False, argtype = 'int'):
        """ parse the user string str for fields definition to store
        into self.attr """

        def _getarg(arg, argtype):
            """ return arg depending on its type """
            if argtype == 'int':
                # arg is the target column index
                try:
                    arg = int(arg)
                except ValueError, e:
                    raise PGLoader_Error, e
                        
            elif argtype == 'char':
                # arg is an escape char
                if len(arg) > 1:
                    raise PGLoader_Error, 'more than one character for char'

            elif argtype == 'string':
                # accept all inputs
                pass

            return arg

        f = self.__dict__[attr] = []

        try:
            serial = 1
            
            for field_def in str.split(','):
                if argtype == 'int' and field_def.find(':') == -1:
                    # support for automatic ordering
                    properties = [field_def.strip(), serial]
                else:
                    properties = [x.strip() for x in field_def.split(':')]

                if not btype:
                    # normal column definition, for COPY usage
                    colname, arg = properties
                    f.append((colname, _getarg(arg, argtype)))

                else:
                    # blob column definition, with blob type, for
                    # UPDATE usage
                    colname, arg, btype = properties
                    f.append((colname, _getarg(arg, argtype), btype))

                # update serial
                if argtype == 'int':
                    serial = int(arg) + 1
                    
        except Exception, error:
            # FIXME: make some errors and write some error messages
            raise
            
    def summary(self):
        """ return a (duration, updates, errors) tuple """
        self.duration = time.time() - self.init_time
        
        if self.reject is not None:
            self.errors = self.reject.errors

        if self.db is not None:
            self.updates = self.db.commited_rows

        return (self.duration, self.updates, self.errors)

    def print_stats(self):
        """ print out some statistics """
        if self.reject is not None:
            self.errors = self.reject.errors
            self.reject.print_stats(self.name)

        if self.db is not None:
            self.updates = self.db.commited_rows
            self.db.print_stats()
        return

    def run(self):
        """ controling thread which dispatch the job """

        # care about number of threads launched
        self.log.debug("%s.sem.acquire()" % self.logname)
        self.sem.acquire()
        self.log.debug("%s acquired starting semaphore" % self.logname)

        # tell parent thread we are running now
        self.started.set()
        self.init_time = time.time()        
        
        try:
            # postinit (will call db.reset() which will get us connected)
            self.log.debug("%s postinit" % self.logname)
            self._postinit()

	    # do the actual processing in do_run and handle all exceptions
            self.do_run()
            
        except Exception, e:
            self.log.error(e)
            self.terminate(False)
            return

        self.terminate()
        return

    def do_run(self):

        # Announce the beginning of the work
        self.log.info("%s processing" % self.logname)

        if self.section_threads == 1:
            # when "No space left on device" where logs are sent,
            # we want to catch the exception
            if 'reader' in self.__dict__ and self.reader.start is not None:
                self.log.debug("Loading from offset %d to %d" \
                               %  (self.reader.start, self.reader.end))

            self.prepare_processing()
            self.process()
            self.finish_processing()

            return

        # Mutli-Threaded processing of current section
        # We want a common Lock() protected Reject object
        #
        # note: this will be done only once, children threads are
        # started with self.section_threads == 1
        self.reject.set_lock(threading.Lock())

        if self.split_file_reading:
            # start self.section_threads workers
            self.split_file_read()

        else:
            # here we need a special thread reading the file
            self.round_robin_read()

        return

    def terminate(self, success = True):
        """ Announce it's over and free the concurrency control semaphore """

        # force PostgreSQL connection closing, do not wait for garbage
        # collector
        if not DRY_RUN:
            self.db.close()

        try:
            self.log.info("releasing %s semaphore" % self.logname)
        except IOError, e:
            # ignore "No space left on device" or other errors here
            pass
        
        self.sem.release()
        
        # tell parent thread processing is now over, here
        try:
            self.log.info("Announce it's over")
        except IOError, e:
            pass
        
        self.success = success
        self.finished.set()
        return

    def split_file_read(self):
        """ Current thread will start self.section_threads threads,
        each one reading a part of the input file. """
        
        # init boundaries to give to each thread
        from stat import ST_SIZE
        previous   = 0
        filesize   = os.stat(self.filename)[ST_SIZE]
        boundaries = []
        for partn in range(self.section_threads):
            start = previous
            end   = (partn+1)*filesize / self.section_threads
            boundaries.append((start, end))

            previous = end + 1

        self.log.debug("Spliting input file of %d bytes %s" \
                      % (filesize, str(boundaries)))

        # Now check for real boundaries
        fd = file(self.filename)
        b  = 0
        for b in range(len(boundaries)):
            start, end = boundaries[b]
            fd.seek(end)
            dummy_str = fd.readline()

            # update both current boundary end and next start
            boundaries[b] = (start, fd.tell()-1)
            if (b+1) < len(boundaries):
                boundaries[b+1] = (fd.tell(), boundaries[b+1][1])

        fd.close()

        self.log.info("Spliting input file of %d bytes %s" \
                      % (filesize, str(boundaries)))

        self.prepare_processing()

        # now create self.section_threads PGLoader threads
        sem      = threading.BoundedSemaphore(self.section_threads)
        summary  = {}
        threads  = {}
        started  = {}
        finished = {}

        for current in range(self.section_threads):
            try:
                summary[current]  = []
                started[current]  = threading.Event()
                finished[current] = threading.Event()
                current_name      = "%s.%d" % (self.name, current)

                loader = PGLoader(self.name, self.config, sem,
                                  (started[current], finished[current]),
                                  summary[current],
                                  logname = current_name,
                                  reject  = self.reject)

                loader.section_threads = 1
                loader.reader.set_boundaries(boundaries[current])
                loader.dont_prepare_nor_finish = True

                threads[current_name] = loader
                threads[current_name].start()

            except Exception, e:
                raise

        # wait for workers to have started, then wait for them to terminate
        check_events(started, self.log, "is running")
        check_events(finished, self.log, "processing is over")
        
        self.finish_processing()
        self.duration = time.time() - self.init_time
        self.log.debug('No more threads are running, %s done' % self.name)

        stats = [0, 0]
        for s in summary:
            for i in range(2, len(summary[s])):
                stats[i-2] += summary[s][i]

        for x in [self.table, self.duration] + stats:
            self.stats.append(x)

        return

    def round_robin_read(self):
        """ Start self.section_threads threads to process data, this
        thread will read the input file and distribute the processing
        on a round-robin fashion"""
        self.prepare_processing()

        try:
            from RRRtools import RRReader
        except ImportError, e:
            raise PGLoader_Error, \
                  "Please upgrade to python 2.4 or newer: %s" % e
        
        queues   = {}
        locks    = {}
        sem      = threading.BoundedSemaphore(self.section_threads)
        summary  = {}
        threads  = {}
        started  = {}
        finished = {}

        for current in range(self.section_threads):
            queues[current] = RRReader()
            locks [current] = threading.Lock()

            # acquire the lock before starting worker thread
            # and release it once its queue if full
            self.log.debug("locks[%d].acquire" % current)
            locks[current].acquire()

            try:
                summary [current] = []
                started [current] = threading.Event()
                finished[current] = threading.Event()
                current_name      = "%s.%d" % (self.name, current)

                loader = PGLoader(self.name, self.config, sem,
                                  (started[current], finished[current]),
                                  summary[current],
                                  logname  = current_name,
                                  reject   = self.reject,
                                  queue    = queues[current],
                                  lock     = locks [current],
                                  copy_sep = self.db.copy_sep)

                loader.section_threads = 1
                loader.dont_prepare_nor_finish = True
                loader.done = False

                threads[current_name] = loader
                threads[current_name].start()

            except Exception, e:
                self.log.error("Couldn't start all workers thread")
                self.log.error(e)

        if len(threads) != self.section_threads:
            self.log.error("Couldn't start all threads, check previous errors")

            check_events([x for x in finished if threads[x].isAlive()],
                         self.log, "processing is over")            
            return

        check_events(started, self.log, "is running")
        
        # Now self.section_threads are started and we have a queue and
        # a Condition for each of them.
        #
        # read the input file here, and give each worker Thread is
        # share to process, in a round-robin fashion
        n = 0               # line number
        c = 0               # current
        p = c               # previous
        
        for offsets, line, columns in self.reader.readlines():
            if p != c:
                self.log.debug("read %d lines, queue to thread %s" % (n, c))

                # release p'thread (which will empty its queue) and
                # lock c'thread --- waiting until it has emptied its
                # queue
                self.log.debug("locks[%d].release" % p)
                locks[p].release()

                # only acquire next thread lock if it's not its first usage
                # cause we acquire() the lock at init time
                #
                # c is 0..self.section_threads, each thread process a
                # queue of self.rrqueue_size elements
                if n > self.rrqueue_size * (c+1):
                    self.log.debug("locks[%d].acquire" % c)
                    locks[c].acquire()
                
            queues[c].append((offsets, line, columns))
            n += 1
            p = c
            c = (n / self.rrqueue_size) % self.section_threads

        # save this for later reference
        last_p = p
        last_c = c

        # we could have some locks to release here
        self.log.debug("p=%d c=%d n=%d (n/rrqueue_size)%%%d=%d " \
                       % (p, c, n,
                          self.section_threads,
                          (n/self.rrqueue_size) % self.section_threads) + \
                       "(n+1/rrqueue_size)%%%d=%d" \
                       % (self.section_threads,
                          ((n+1)/self.rrqueue_size) % self.section_threads))
        
        if p != c or (n % self.rrqueue_size != 0):
            self.log.debug("locks[%d].release" % p)
            locks[p].release()

        # we could have read all the data and not needed all workers,
        # log it when it's the case, then set .done = True without
        # taking again a lock which we already have.
        if n < (self.section_threads * self.rrqueue_size):
            self.log.info("processed all data with only %d workers" % (c+1))

        # mark all worker threads has done
        k = threads.keys()
        for c in range(self.section_threads):
            # we don't need any lock.acquire if we didn't use the worker
            if n > (self.section_threads * self.rrqueue_size) \
               or c <= last_c:
                self.log.debug("locks[%d].acquire to set %s.done = True" \
                           % (c, k[c]))            
                locks[c].acquire()

            
            threads[k[c]].done = True
            
            self.log.debug("locks[%d].release (done set)" % c)
            locks[c].release()

        # wait for workers to finish processing
        check_events(finished, self.log,  "processing is over")
        
        self.finish_processing()
        self.duration = time.time() - self.init_time
        self.log.debug('%s done' % self.name)

        stats = [0, 0]
        for s in summary:
            for i in range(2, len(summary[s])):
                stats[i-2] += summary[s][i]

        for x in [self.table, self.duration] + stats:
            self.stats.append(x)

        return

    def readlines(self):
        """ return next line from either self.queue or self.reader """

        if self.queue is None:
            for offsets, line, columns in self.reader.readlines():
                yield offsets, line, columns

            return

        while not self.done:
            self.lock.acquire()

            if len(self.queue) > 0:
                self.log.debug("processing queue")
                for offsets, line, columns in self.queue.readlines():
                    yield offsets, line, columns

            self.lock.release()

        return

    def prepare_processing(self):
        """ Things to do before processing data """
        if 'dont_prepare_nor_finish' in self.__dict__:
            return
        
        if not DRY_RUN:
            if TRUNCATE:
                self.db.truncate(self.table)
                
            if TRIGGERS:
                self.db.disable_triggers(self.table)

    def finish_processing(self):
        """ Things to do after processing data """
        if 'dont_prepare_nor_finish' in self.__dict__:
            return

        if TRIGGERS and not DRY_RUN:
            self.db.enable_triggers(self.table)

        self.log.debug("loading done")
        return

    def update_summary(self):
        """ update the main summary """
        self.duration = time.time() - self.init_time
        
        if self.reject is not None:
            self.errors = self.reject.errors

        if DRY_RUN:
            self.commited_rows = 0
        else:
            self.commited_rows = self.db.commited_rows
            
        for x in [self.table, self.duration,
                  self.commited_rows, self.errors]:
            self.stats.append(x)

        # then show up some stats
        self.print_stats()
        
    def process(self):
        """ depending on configuration, do given job """

        if self.columns is not None:
            self.data_import()

        elif self.blob_cols is not None:
            # elif: COPY process also blob data
            self.log.info("UPDATE blob data")

        self.update_summary()

    def data_import(self):
        """ import CSV or TEXT data, using COPY """

        # some more practical data format of internals
        ddict = dict(self.columns)
        if self.reformat:
            drefc = dict(self.reformat)
                
        if self.udcs:
            dudcs = dict(self.udcs)

        for offsets, line, columns in self.readlines():
            self.log.debug('offsets %s', offsets)

            if self.blob_cols is not None:
                columns, rowids = self.read_blob(line, columns)

            if self.reformat:
                refc = dict(self.reformat)
                data = []
                for cname, cpos in self.columns:
                    if cname in drefc:
                        # reformat the column value
                        data.append(drefc[cname](self.reject,
                                                 columns[cpos-1]))
                    else:
                        data.append(columns[cpos-1])

                self.log.debug('reformat')
                self.log.debug('columns %s', str(columns))
                self.log.debug('data    %s', str(data))

                # we want next steps to take reformated data as input
                columns = data
                    
            if self.udcs:
                dudcs = dict(self.udcs)
                data = []
                for c in self.copy_columns:
                    if c in ddict:
                        data.append(columns[ddict[c]-1])
                    else:
                        data.append(dudcs[c])

                self.log.debug('udcs')
                self.log.debug('columns %s', str(columns))
                self.log.debug('data    %s', str(data))
                
                columns = data
                
            else:
                if self.col_mapping:
                    self.log.debug('col_mapping %s', str(self.col_mapping))

                    data = [columns[i-1] for i in self.col_mapping]

                    self.log.debug('columns %s', str(columns))
                    self.log.debug('data    %s', str(data))

                    columns = data

            if self.only_cols:
                data = [columns[i-1] for i in self.only_cols]

            if not self.reformat \
                   and not self.udcs \
                   and not self.col_mapping \
                   and not self.only_cols:
                data = columns

            if DRY_RUN:
                self.log.info('< %s', line)
                self.log.info('  %s', str(self.columnlist))
                self.log.info('> %s', str(data))
            else:
                self.log.debug('< %s', line)
                self.log.debug('  %s', str(self.columnlist))
                self.log.debug('> %s', str(data))
                    
            if not DRY_RUN:
                self.db.copy_from(self.table, self.columnlist,
                                  data, line, offsets, self.reject)

        if not DRY_RUN:
            # we may need a last COPY for the rest of data
            self.db.copy_from(self.table, self.columnlist,
                              None, None, None, self.reject, EOF = True)

        return

    ##
    # BLOB data reading/parsing
    # probably should be moved out from this file

    def lo_import(self):
        """ import large object data, using UPDATEs """

        for line, columns in self.read_data():
            ##
            # read blob data and replace its PG escaped form into columns
            columns, rowids = self.read_blob(line, columns)
            
            # and insert it into database
            self.log.debug(line)
            self.log.debug(str(line))

            if not DRY_RUN:
                self.db.insert_blob(self.table, self.index,
                                    rowids, cname, data, btype,
                                    line, self.reject)



    def read_blob(self, line, columns):
        """ read blob data and replace columns values with PG escape
        data to give as input to COPY or UPDATE """

        # cache rowids, a single line ids will be the same whatever
        # the number of clob entries are configured
        rowids = None

        for (cname, c, btype) in self.blob_cols:
            try:
                blob_col = c-1
                blob = columns[blob_col].strip()
            except Exception, e:
                m = [
                    "Warning: couldn't get '%s' column (#%d)" % (cname, c),
                    str(e)
                    ]
                try:
                    self.reject.log(m, line)
                except PGLoader_Error:
                    # pedantic setting, but we want to continue
                    continue

            if blob != "":
                try:
                    (begin, length, blobname) = blob.split(CLOB_SEP)

                except Exception, e:
                    m = ["Warning: column '%s'" % cname + \
                         " is not a valid blob reference: %s" % blob,
                         str(e)
                         ]
                    self.reject.log(m, line)
                    continue

                # Informix sometimes output 0,0,0 as blob reference
                if begin == length == blobname == "0":
                    columns[blob_col] = ''
                    continue

                abs_blobname = os.path.join(os.path.dirname(self.filename),
                                            blobname)

                # We want to manage existing blob file
                if not os.access(abs_blobname, os.R_OK):
                     self.reject.log(
                         "Warning: Can't read blob file %s" % abs_blobname,
                         line)

                if abs_blobname not in self.blobs:

                    if btype not in ['ifx_blob', 'ifx_clob']:
                        msg = "Error: Blob type '%s' not supported" % mode
                        raise PGLoader_Error, msg

                    elif btype == 'ifx_blob':
                        self.blobs[abs_blobname] = ifx_blob(self.log,
                                                            abs_blobname, 
                                                            self.field_sep)

                    elif btype == 'ifx_clob':
                        self.blobs[abs_blobname] = \
                                                 ifx_clob(self.log,
                                                          abs_blobname,
                                                          self.input_encoding)

                blob = self.blobs[abs_blobname]

                # we now need to get row id(s) from ordered columns
                if rowids is None:
                    rowids = self._rowids(columns)

                # get data from blob object
                data = blob.extract(rowids, c, begin, length)
                columns[blob_col] = data

                self.log.debug('read blob data')
                self.log.debug(str(rowids))
                self.log.debug(str(data))

        # no we have read all defined blob data, returns new columns
        return columns, rowids
                                           
                    
    def _rowids(self, columns):
        """ get rowids for given input line """
        rowids = {}
        try:
            for id_name, id_col in self.index:
                rowids[id_name] = columns[id_col - 1]

        except IndexError, e:
            messages = [
                "Warning: couldn't get id %d on column #%d" % (id_name,
                                                               id_col), 
                str(e)
                ]
                
            self.reject.log(messages, line)

        return rowids




# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader text format reader
#
# handles configuration, parse data, then pass them to database module for
# COPY preparation

import os, sys, os.path, time

from tools    import PGLoader_Error, Reject, parse_config_string
from db       import db
from reader   import DataReader, UnbufferedFileReader

from options import DRY_RUN, PEDANTIC
from options import TRUNCATE, VACUUM
from options import COUNT, FROM_COUNT, FROM_ID
from options import INPUT_ENCODING, PG_CLIENT_ENCODING
from options import COPY_SEP, FIELD_SEP, CLOB_SEP, NULL, EMPTY_STRING
from options import NEWLINE_ESCAPES

class FixedReader(DataReader):
    """
    Read fixed file format, configuration gives for each field
     - field name
     - start position
     - length
    """

    def readconfig(self, config, name, template):
        """ get this reader module configuration from config file """
        DataReader.readconfig(self, config, name, template)

        # this will be called twice if templates are in used, so we
        # have to protect ourselves against removing already read
        # configurations while in second run.

        self._getopt('fixed_specs', config, name, template, None)

        if self.fixed_specs:
            self.positions = {}
            # parse the fixed specs
            specs = [x.strip().split(':') for x in self.fixed_specs.strip().split(',')]
            try:
                for name, start, length in specs:
                    self.positions[name] = (int(start), int(length))
                    
            except ValueError, e:
                self.log.error("%s.fixed_specs, " + \
                               "start and length must be numbers", name)
                
                raise PGLoader_Error, \
                      "Please fix %s.fixed_specs configuration" % name
        else:
            msg = "section %s: fixed format type require 'fixed_specs'" % name
            raise PGLoader_Error, msg

        self.log.debug('reader.readconfig: positions %s', self.positions)

    def readlines(self):
        """ read data from configured file, and generate (yields) for
        each data line: line, columns and rowid """

        self.fd = UnbufferedFileReader(self.filename, self.log,
                                       encoding = self.input_encoding,
                                       start    = self.start,
                                       end      = self.end,
                                       skip_head_lines = self.skip_head_lines)

        # don't forget COUNT and FROM_COUNT option
        nb_lines     = self.skip_head_lines
        begin_linenb = None

        for line in self.fd:
            line      = line.strip("\n")
            llen      = len(line)
            columns   = []
            offsets   = [self.fd.line_nb]
            nb_lines += 1

            for cname, cpos in self.columns:
                start, length = self.positions[cname]

                if llen < (start+length):
                    self.log.error("Line %d is too short " % self.fd.line_nb +
                                   "(column %s requires len >= %d)" \
                                   % (cname, start+length))

                    msg = "Please review fixed_specs configuration"
                    raise PGLoader_Error, msg
                
                columns.append(line[start:start+length])

            if self.start:
                offsets = (self.start, offsets)

            yield offsets, line, columns



# Author: Dimitri Fontaine <dim@tapoueh.org>
#
# pgloader librairies

import os, sys, os.path, time, codecs
from cStringIO import StringIO

from options import DRY_RUN, PEDANTIC

class PGLoader_Error(Exception):
    """ Internal pgloader processing error """
    pass

class Reject:
    """ We log rejects into two files, reject_log and reject_data

    reject_log  contains some error messages and reasons
    reject_data contains input lines which this tool couldn't manage
    """

    def __init__(self, log, reject_log, reject_data):
        """ Constructor, with file names """
        self._log        = log
        self.reject_log  = reject_log
        self.reject_data = reject_data
        self.lock        = None

        # we will open files on first error
        self.errors = 0

    def set_lock(self, lock):
        """ when used in a multi-threaded way, you want a common
        reject facility, protected from concurrent writes """
        self.lock = lock

    def print_stats(self, name):
        """ give a summary """
        if DRY_RUN:
            return
        
        if self.errors == 0:
            self._log.info("No data were rejected")
        else:
            self._log.error("%d errors found into [%s] data",
                            self.errors, name)
            self._log.error("please read %s for errors log", self.reject_log)
            self._log.error("and %s for data still to process",
                            self.reject_data)

    def log(self, messages, data = None):
        """ Acquire the lock if needed, do_log() and check for IOError """

        if self.lock:
            self._log.debug("Reject acquire")
            self.lock.acquire()

        try:
            self.do_log(messages, data)

        except IOError, e:
            raise PGLoader_Error, e

        except PGLoader_Error, e:
            raise 

        if self.lock:
            self._log.debug("Reject release")
            self.lock.release()

    def do_log(self, messages, data = None):
        """ log the messages into reject_log, and the data into reject_data

        We open the file on each request, cause we supose errors to be
        rare while the import process will take a long time.
        """

        if self.errors == 0:
            fd_log  = open(self.reject_log,  'wb+')
            fd_data = open(self.reject_data, 'wb+')
        else:
            fd_log  = open(self.reject_log,  'ab+')
            fd_data = open(self.reject_data, 'ab+')

        # message has to be either a string or a list of strings
        if type(messages) == type("string"):
            error = messages + "\n"
            fd_log.write(error)

            if PEDANTIC and not VERBOSE:
                # (write the message just once)
                sys.stderr.write(error)
            
        else:
            error = None
            for m in messages:
                if error is None: error = m
                m += "\n"
                fd_log.write(m)
                
                if PEDANTIC:
                    sys.stderr.write(m)
                    
        # add a separation line between log entries
        fd_log.write("\n")

        # data has to be a string, a single input line or None
        # the input line is not chomped
        if data is not None:
            fd_data.write(data)

        # now we close the two fds
        for f in [fd_log, fd_data]:
            f.flush()
            f.close()

        self.errors += 1

        if PEDANTIC:
            raise PGLoader_Error, error


def parse_config_string(str):
    """ parse a config string

    used for null and empty_string elements
    null = ""
    empty_string = "\ "

    this would result in null =="" and empty_string == '"\ "',
    which is not what we want.
    """

    if len(str) > 2:
        if (str[0] == str[-1] == '"') \
               or (str[0] == str[-1] == "'"):
            # we have a param = "foo" configuration, we want to return only
            # the foo
            return str[1:-1]

    return str


def parse_pg_options(log, config, section, pg_options, overwrite=False):
    """ Get all the pg_options_ prefixed options from the section"""
    # PostgreSQL options must begin with the prefix pg_option_
    for o in [x for x in config.options(section) 
              if x.startswith('pg_option_')]:
        opt = o[len('pg_option_'):]
        val = config.get(section, o)

        # hysterical raisins
        for compat in ['client_encoding', 'lc_messages', 'datestyle']:
            if opt == compat and config.has_option(section, compat):
                log.warning("Ignoring %s.%s for %s.%s" \
                                % (section, o, section, opt))

        if opt not in compat and (overwrite or opt not in pg_options):
            pg_options[opt] = val

    return pg_options
    
def read_path(strpath, log, path = [], check = True):
    """ read a path configuration element, discarding non-existing entries """
    import os.path

    for p in strpath.split(':'):
        path.append(p)

    if check:
        return check_path(path, log)
    else:
        return path

def check_path(path, log):
    """ removes non existant and non {directories, symlink} entries from path
    """
    path_ok = []

    for p in path:
        if os.path.exists(p):
            if os.path.isdir(p) or \
                   (os.path.islink(p) and os.path.isdir(os.path.realpath(p))):
                path_ok.append(p)
            else:
                log.warning("path entry '%s' is not a directory " + \
                            "or does not link to a directory", p)
        else:
            log.warning("path entry '%s' does not exists, ignored" % p)

    return path_ok


def check_dirname(path):
    """ check if given path dirname exists, try to create it if if doesn't """

    # try to create the log file and the directory where it lives
    logdir = os.path.dirname(path)
    if logdir and not os.path.exists(logdir):
        # logdir is not empty (not CWD) and does not exists
        try:
            os.makedirs(logdir)
        except (IOError, OSError), e:
            return False, e

    return True, None


def check_events(events, log, context = "is running"):
    """ wait until all events (list) are set """
    for t in events:
        events[t].wait()
        log.debug("thread %s %s" % (t, context))
    
    return
